from guizero import App, Picture, PushButton, info
app = App()

anim = Picture(app, "guizero_flash.gif")
anim.width = 200
anim.height = 100

button = PushButton(app, command=info, args=["button", "you pushed the image"], image="guizero_flash.gif")
button.width = 200
button.height = 100

app.display()


from threading import Event
from guizero import App, ListBox
from common_test import (
    schedule_after_test,
    schedule_repeat_test,
    destroy_test,
    enable_test,
    display_test,
    text_test,
    color_test,
    size_text_test,
    size_fill_test,
    events_test,
    cascaded_properties_test,
    inherited_properties_test,
    grid_layout_test,
    auto_layout_test
    )

def test_default_values():
    a = App()
    l = ListBox(a)
    assert l.master == a
    assert l.value == None
    assert l.items == []
    assert l.grid == None
    assert l.align == None
    assert l.visible == True
    assert l.enabled == True
    assert l._listbox._multiselect == False
    assert a.description > ""
    a.destroy()

def test_alt_values():
    a = App(layout = "grid")
    l = ListBox(
        a,
        ["foo", "bar"],
        selected = "bar",
        grid = [0,1],
        align = "top",
        width=10,
        height=11,
        scrollbar=True)

    assert l.value == "bar"
    assert l.items == ["foo", "bar"]
    assert l.grid[0] == 0
    assert l.grid[1] == 1
    assert l.align == "top"
    assert l.width == 10
    assert l.height == 11

    a.destroy()

def test_multi_default_values():
    a = App()
    l = ListBox(a, multiselect = True)
    assert l.master == a
    assert l.value == None
    assert l.items == []
    assert l.grid == None
    assert l.align == None
    assert l.visible == True
    assert l.enabled == True
    assert l._listbox._multiselect == True
    a.destroy()

def test_multi_alt_values():
    a = App(layout = "grid")
    l = ListBox(
        a,
        ["foo", "bar"],
        selected = ["bar"],
        grid = [0,1],
        align = "top",
        multiselect = True,
        scrollbar=True)

    assert l.value == ["bar"]
    assert l.items == ["foo", "bar"]
    assert l.grid[0] == 0
    assert l.grid[1] == 1
    assert l.align == "top"
    assert l._listbox._multiselect == True
    a.destroy()

def test_getters_setters():
    a = App()
    l = ListBox(a, ["foo", "bar"])

    assert l.value == None
    l.value = "bar"
    assert l.value == "bar"

    a.destroy()

def test_multi_getters_setters():
    a = App()
    l = ListBox(a, ["foo", "bar"], multiselect=True)

    assert l.value == None
    l.value = ["bar"]
    assert l.value == ["bar"]
    l.value = ["bar", "foo"]
    # test selected values are in value
    assert all(x in l.value for x in ['bar', 'foo'])

    a.destroy()

def test_append():
    a = App()
    l = ListBox(a, ["foo", "bar"])

    assert l.items == ["foo", "bar"]
    l.append("car")
    assert l.items == ["foo", "bar", "car"]

    a.destroy()

def test_insert():
    a = App()
    l = ListBox(a, ["foo", "bar"])

    assert l.items == ["foo", "bar"]
    l.insert(1, "car")
    assert l.items == ["foo", "car", "bar"]

    a.destroy()

def test_remove():
    a = App()
    l = ListBox(a, ["foo", "bar", "foo"])

    assert l.items == ["foo", "bar", "foo"]
    l.remove("foo")
    assert l.items == ["bar", "foo"]

    a.destroy()

def test_clear():
    a = App()
    l = ListBox(a, ["foo", "bar"])
    assert l.items == ["foo", "bar"]
    l.clear()
    assert l.items == []

    a.destroy()

def test_command():
    a = App()

    callback_event = Event()
    def callback():
        callback_event.set()

    l = ListBox(a, ["foo", "bar"], command = callback)
    assert not callback_event.is_set()

    l._listbox._command_callback()
    assert callback_event.is_set()

    a.destroy()

def test_command_with_parameter():
    a = App()

    callback_event = Event()
    def callback(value):
        assert value == "bar"
        callback_event.set()

    l = ListBox(a, ["foo", "bar"], command = callback)
    l.value = "bar"
    assert not callback_event.is_set()

    l._listbox._command_callback()
    assert callback_event.is_set()

    a.destroy()

def test_update_command():
    a = App()

    callback_event = Event()
    def callback():
        callback_event.set()

    l = ListBox(a, ["foo", "bar"])

    l._listbox._command_callback()
    assert not callback_event.is_set()

    l.update_command(callback)
    l._listbox._command_callback()
    assert callback_event.is_set()
    callback_event.clear()

    l.update_command(None)
    l._listbox._command_callback()
    assert not callback_event.is_set()

    a.destroy()

def test_update_command_with_parameter():
    a = App()

    callback_event = Event()
    def callback(value):
        assert l.value == "foo"
        callback_event.set()

    l = ListBox(a, ["foo", "bar"])
    l.value = "foo"

    l.update_command(callback)

    l._listbox._command_callback()
    assert callback_event.is_set()

    a.destroy()

def test_after_schedule():
    a = App()
    l = ListBox(a, ["foo", "bar"])
    schedule_after_test(a, l)
    a.destroy()

def test_repeat_schedule():
    a = App()
    l = ListBox(a, ["foo", "bar"])
    schedule_repeat_test(a, l)
    a.destroy()

def test_destroy():
    a = App()
    l = ListBox(a, ["foo", "bar"])
    destroy_test(l)
    a.destroy()

def test_enable():
    a = App()
    l = ListBox(a, ["foo", "bar"])
    enable_test(l)
    a.destroy()

def test_display():
    a = App()
    l = ListBox(a, ["foo", "bar"])
    display_test(l)
    a.destroy()

def test_text():
    a = App()
    l = ListBox(a, ["foo", "bar"])
    text_test(l)
    a.destroy()

def test_color():
    a = App()
    l = ListBox(a, ["foo", "bar"])
    color_test(l)
    a.destroy()

def test_size():
    a = App()
    l = ListBox(a, ["foo", "bar"])
    size_text_test(l)
    size_fill_test(l)
    a.destroy()

def test_events():
    a = App()
    l = ListBox(a, ["foo", "bar"])
    events_test(l)
    a.destroy()

def test_cascaded_properties():
    a = App()
    l = ListBox(a, ["foo", "bar"])
    cascaded_properties_test(a, l, True)
    a.destroy()

def test_inherited_properties():
    a = App()
    inherited_properties_test(a, lambda: ListBox(a, ["foo", "bar"]), True)
    a.destroy()

def test_auto_layout():
    a = App()
    w = ListBox(a)
    auto_layout_test(w, None)
    a.destroy()

def test_grid_layout():
    a = App(layout="grid")
    
    w = ListBox(a, grid=[1,2])
    grid_layout_test(w, 1, 2, 1, 1, None)
    
    ws = ListBox(a, grid=[1,2,3,4])
    grid_layout_test(ws, 1, 2, 3, 4, None)

    wa = ListBox(a, grid=[1,2], align="top")
    grid_layout_test(wa, 1, 2, 1, 1, "top")
    
    a.destroy()

test_multi_alt_values()



from guizero import App, Box, ButtonGroup, CheckBox, Combo, PushButton, Slider, Text, TextBox

a = App(title="colors")
text = Text(a, text="colors")
check = CheckBox(a, "check me")
combo = Combo(a, ["red", "blue"])
button = PushButton(a)
slider = Slider(a)
b = Box(a)
textbox = TextBox(b, text="or colours")
bgroup = ButtonGroup(b, ["cheese", "ham", "salad"], 1)

#a.bg = (255,255,0)
text.text_color = "red"
text.text_size = 30
text.font = "verdana"
#text.bg = "green"
check.bg = "#d41789"
combo.bg = "blue"
combo.text_color = None
combo.text_size = 24
#button.bg = "black"
button.text_color = (255,0,255)
button.font = "arial"
button.text_size = 18
slider.bg = (123,234,12)
#textbox.bg = "cyan"
textbox.font = "courier"
textbox.text_color = "#FF0000"
b.bg = "cyan"
b.font = "wingdings"
#bgroup.bg = "yellow"
bgroup.text_color = "#e62112"
bgroup.text_color = None
bgroup.font = "book antiqua"

#a.bg = (255,0,25)
#a.font = "wingdings"
a.text_color = (255,253,12)
b.text_color = "purple"
b.text_size = 18

a.display()


from tkinter import Tk, Toplevel
from .base import BaseWindow
from . import utilities as utils, system_config


class App(BaseWindow):

    _main_app = None

    def __init__(
        self, 
        title="guizero", 
        width=500, 
        height=500, 
        layout="auto", 
        bg=None, 
        visible=True):

        """
        Creates an App

        :param string title:
            The text in the title bar of the window, defaults to `guizero`.

        :param int width:
            The width of the App window, defaults to 500.

        :param int height:
            The height of the App window, defaults to 500.

        :param string layout:
            The layout manager style for this window, defaults to `auto`.

        :param color bg:
            The background colour for this window, defaults to None. See https://lawsie.github.io/guizero/colors/

        :param bool visible:
            If the widget should be visible, defaults to `True`.

        """

        # If this is the first app to be created, create Tk
        if App._main_app is None:
            tk = Tk()
            App._main_app = self
            # apply tk options
            for option_key in system_config.tk_options:
                tk.option_add(option_key, system_config.tk_options[option_key])
        else:
            tk = Toplevel(App._main_app.tk)
            utils.error_format("There should only be 1 guizero App, use Window to create multiple windows.")

        super(App, self).__init__(
            None,
            tk,
            title,
            width,
            height,
            layout,
            bg,
            visible)

    # METHODS
    # --------------------------------------


    def display(self):
        """
        Display the window.

        :return:
            None.
        """
        self.tk.mainloop()

    def destroy(self):
        """
        Destroy and close the App.

        :return:
            None.

         :note:
            Once destroyed an App can no longer be used.
        """
        # if this is the main_app - set the _main_app class variable to `None`.
        if self == App._main_app:
            App._main_app = None
        self.tk.destroy()



from guizero import App, Drawing

a = App(width = 370, height=700)

# create drawing object
d = Drawing(a, width="fill", height="fill")
d.bg = "light blue"

# draw the shapes
d.rectangle(10, 10, 60, 60)
d.rectangle(70, 10, 120, 60, color="yellow")
d.rectangle(130, 10, 180, 60, color="yellow", outline=True)
d.rectangle(190, 10, 240, 60, color="yellow", outline=5)
d.rectangle(250, 10, 300, 60, color="yellow", outline=5, outline_color="green")
d.rectangle(310, 10, 360, 60, color=None, outline=5, outline_color="red")

d.oval(10, 100, 60, 150)
d.oval(70, 100, 120, 200, color="yellow")
d.oval(130, 100, 240, 150, color="yellow", outline=True)
d.oval(130, 160, 240, 210, color="yellow", outline=5)
d.oval(250, 100, 300, 150, color="yellow", outline=5, outline_color="green")
d.oval(310, 100, 360, 150, color=None, outline=5, outline_color="red")

d.line(10, 250, 60, 250)
d.line(70, 250, 120, 250, color="yellow")
d.line(130, 250, 240, 250, width=5)
d.line(250, 250, 300, 250, width=5, color="green")
d.line(310, 250, 360, 250, width=5, color="red")

d.polygon(10, 300, 60, 300, 40, 350, 10, 350)
d.polygon(70, 300, 120, 300, 100, 350, 70, 350, color="yellow")
d.polygon(130, 300, 180, 300, 160, 350, 130, 350, color="yellow", outline=True)
d.polygon(190, 300, 240, 300, 220, 350, 190, 350, color="yellow", outline=5)
d.polygon(250, 300, 300, 300, 280, 350, 250, 350, color="yellow", outline=5, outline_color="green")
d.polygon(310, 300, 360, 300, 340, 350, 310, 350, color=None, outline=5, outline_color="green")

d.triangle(10, 400, 60, 400, 10, 450)
d.triangle(70, 400, 120, 400, 70, 450, color="yellow")
d.triangle(130, 400, 180, 400, 130, 450, color="yellow", outline=True)
d.triangle(190, 400, 240, 400, 190, 450, color="yellow", outline=5)
d.triangle(250, 400, 300, 400, 250, 450, color="yellow", outline=5, outline_color="green")
d.triangle(310, 400, 360, 400, 310, 450, color=None, outline=5, outline_color="green")

d.image(10, 500, "guizero.png", width=350, height=100)

d.text(10, 600, "guizero")
d.text(110, 600, "guizero", font="times new roman")
d.text(210, 600, "guizero", size=24)
d.text(10, 650, "this is a some text which goes over the width and is wrapped", font="arial", size=16, max_width=350)

a.display()



from guizero import App, Text

def do_this_on_close():
    # Ask user if they really want to close the window
    if app.yesno("Close", "Do you want to quit?"):
        app.destroy()

app = App()

title = Text(app, "blank app")

# When user tries to close the window, execute do_this_on_close()
app.when_closed = do_this_on_close

app.display()


from guizero import App, TextBox, PushButton, Text

def show():
    output.value = textbox.value

app = App()
textbox = TextBox(app, multiline=True, height=10, width=50, scrollbar=True)
textbox.value = "hello\ngoodbye\nno way\nthis is a very long stream of text, very long indeed, the best long line of text, its super bigly and very long, I dont think it could possibly be any better particularly as it was created by someone who is super good at creating long lines of text."
button = PushButton(app, text="Print", command=show)
output = Text(app)
app.display()


from guizero import App, Slider, Picture

def resize():
    picture.resize(width.value, height.value)

app = App(layout="grid", width=550, height=200)

picture = Picture(app, image="guizero.gif", grid=[0,1])

width = Slider(app, command=resize, grid=[0,0], start=1, end=picture.width)
width.width = picture.width
width.value = picture.width

height = Slider(app, command=resize, horizontal=False, grid=[1,1], start=1, end=picture.height)
height.height = picture.height
height.value = picture.height

app.display()



from tkinter import Menu
from .tkmixins import ScheduleMixin, DestroyMixin, FocusMixin
from . import utilities as utils
from .base import Component
from .App import App
from .Window import Window

class MenuBar(Component):

    def __init__(self, master, toplevel, options):

        """
        Creates a MenuBar

        :param Container master:
            The Container (App, Box, etc) the MenuBar will belong too.

        :param List toplevel:
            A list of strings to populate the top level menu options.

        :param List options:
            A 3D list of:
                - submenus,
                - with each submenu being a list of options
                - and each option being a text/command pair 

            e.g ::

                options=[
                        [ ["File option 1", file_function], ["File option 2", file_function] ],
                        [ ["Edit option 1", edit_function], ["Edit option 2", edit_function] ]
                    ]
        """

        if not isinstance(master, (App, Window)):
            utils.error_format("The [MenuBar] must have an [App] or [Window] object as its master")

        # Create a tk Menu object within this object
        tk = Menu(master.tk)

        super(MenuBar, self).__init__(master, tk, False)

        # Keep track of submenu objects
        self._sub_menus = []

        # Create all the top level menus
        for i in range(len(toplevel)):
            # Create this submenu
            new_menu = Menu(self.tk, tearoff=0)

            # Populate the drop down menu with the items/commands from the list
            for menu_item in options[i]:
                new_menu.add_command(label=menu_item[0], command=menu_item[1])

            # Append to the submenus list
            self._sub_menus.append(new_menu)

            # Add to the menu bar
            self.tk.add_cascade(label=toplevel[i], menu=self._sub_menus[i])

       	# Set this as the menu for the master object
       	master.tk.config(menu=self.tk)

    @property
    def bg(self):
        """
        Sets the background color of the widget.

        Note - some operating systems dont allow the background color of the 
        menu bar to be changed. 
        """
        return super(MenuBar, self.__class__).bg.fget(self)

    # Set the background colour
    @bg.setter
    def bg(self, color):
        super(MenuBar, self.__class__).bg.fset(self, color)
        for sub_menu in self._sub_menus:
            sub_menu["bg"] = utils.convert_color(color)
    


from tkinter import Button, StringVar, DISABLED, NORMAL
from . import utilities as utils
from .base import TextWidget

class PushButton(TextWidget):
    def __init__(
        self,
        master,
        command=None,
        args=None,
        text="Button",
        image=None,
        pady=10,
        padx=10,
        grid=None,
        align=None,
        icon=None,
        visible=True,
        enabled=None,
        width=None,
        height=None):

        """
        Creates a PushButton

        :param Container master:
            The Container (App, Box, etc) the Picture will belong to.

        :param Callable command:
            A string containing the image to display, defaults to `None`.

        :param List args:
            A list of arguments to pass to the command. Defaults to `None`.

        :param string text:
            The text to display on the button, defaults to `Button`.

        :param string image:
            A string containing the image to display, defaults to `None`.
            If an image is specified, this overrides any text set to display
            on the button.

        :param int padx:
            The amount of horizontal padding the button should have. Defaults to 10.

        :param int pady:
            The amount of vertical padding the button should have. Defaults to 10.

        :param List grid:
            Grid co-ordinates for the widget, required if the master layout
            is 'grid', defaults to `None`.

        :param string align:
            How to align the widget within the grid, defaults to None.

        :param string icon:
            A string containing the image to display, defaults to `None`.
            If an image is specified, this overrides any text set to display
            on the button. (DEPRECATED)

        :param bool visible:
            If the widget should be visible, defaults to `True`.

        :param bool enabled:
            If the widget should be enabled, defaults to `None`. If `None`
            the value is inherited from the master.

        :param int width:
            The starting width of the widget. Defaults to `None` and will auto
            size.

        :param int height:
            The starting height of the widget. Defaults to `None` and will auto
            size.
        """

        self._value = 0
        self._image_source = icon
        self._image_source = image
        self._image = None
        self._image_height = None
        self._image_width = None
        self._image_player = None

        # Create a tk Button object within this object
        self._text = StringVar()
        self._text.set(text)
        tk = Button(master.tk, textvariable=self._text, command=self._command_callback)

        super(PushButton, self).__init__(master, tk, grid, align, visible, enabled, width, height)

        # Add padding if necessary
        self.tk.config(pady=pady, padx=padx)

        # Setup events for press and release
        self.events.set_event("<PushButton.ButtonPress-1>", "<ButtonPress-1>", self._on_press)
        self.events.set_event("<PushButton.ButtonRelease-1>", "<ButtonRelease-1>", self._on_release)

        self.update_command(command, args)

        if image:
            self._load_image()

    def _load_image(self):
        # clear the text on the button if an image is being loaded
        self._text.set("")
        
        if self._height == "fill" or self._width == "fill":
            utils.raise_error("{}\nCannot use 'fill' for width and height when using a image.".format(self.description))

        # stop any animation which might still be playing
        if self._image_player:
            self._image_player.stop()

        self._image = utils.GUIZeroImage(self._image_source, self._width, self._height)

        # update the tk image
        # if its an animation, start it up
        if self._image.animation:
            self._image_player = utils.AnimationPlayer(self, self._image, self._update_tk_image)
        else:
            self._update_tk_image(self._image.tk_image)

        # set the width and height of the widget to match the image if they are None
        super(PushButton, self.__class__).resize(
            self,
            self._image.width if self.width is None else self.width,
            self._image.height if self.height is None else self.height)

    def _update_tk_image(self, tk_image):
        self.tk.config(image=tk_image)

    # PROPERTIES
    # ----------------------------------
    # Get the value of the button (pressed, released)
    @property
    def value(self):
        return self._value

    # Get the text on the button
    @property
    def text(self):
        return (self._text.get())

    # Set the text on the button
    @text.setter
    def text(self, value):
        # clear any existing image from the button
        self._clear_image()
        self._text.set(str(value))
        
    @property
    def image(self):
        return None if self._image is None else self._image.image_source
        
    @image.setter
    def image(self, value):
        self._clear_image()
        if value is not None:
            self._image_source = value
            self._load_image()    

    @property
    def description(self):
        """
        Returns the description for the widget.
        """
        return "[PushButton] object with text '{}'".format(self.text)

    # METHODS
    # -------------------------------------------
    # Internal use only
    # Called when the button is pressed
    def _on_press(self, event):
        if self.enabled:
            self._value = 1

    # Called when the button is released
    def _on_release(self, event):
        if self.enabled:
            self._value = 0

    # Change padding
    def padding(self, padx, pady):
        self.tk.config(padx=padx, pady=pady)

    def toggle(self):
        self.enabled = not self.enabled

    def update_command(self, command, args=None):
        if command is None:
            self._command = lambda: None
        else:
            if args is None:
                self._command = command
            else:
                self._command = utils.with_args(command, *args)
    
    def resize(self, width, height):
        super(PushButton, self.__class__).resize(self, width, height)

        if self._image:
            self._load_image()

    def _command_callback(self):
        self._command()

    def _clear_image(self):
        if self._image:
            self._image_source = None
            self._image = None
            self._image_height = None
            self._image_width = None
            self._image_player = None
            self.tk.config(image="")
            self.resize(None, None)
            self._text.set(str("Button"))


from guizero import App
app = App()
app.info("Info", "This is a guizero app")
app.error("Error", "Try and keep these out your code...")
app.warn("Warning", "These are helpful to alert users")
app.display()


from guizero import App, Box, Text, PushButton, Combo, Slider

def switch_screen(switch_to):
    hide_all()
    switch_to.show()

def hide_all():
    for screen in all_screens:
        screen.hide()


app = App("Multi box app", layout="grid")

# Create a blank list to hold all the different screens
all_screens = []

# Create a box to contain the menu buttons
menu = Box(app, grid=[0,0], layout="grid")
menu.tk.width = 900
menu.bg = "red"

# Option 1 box
option1 = Box(app, grid=[1,1])
text1 = Text(option1, text="This is the first page of stuff")
combo = Combo(option1, options=["Beef", "Chicken", "Fish", "Vegetarian"])
all_screens.append(option1)

# Option 2 box
option2 = Box(app, grid=[1,1])
text2 = Text(option2, text="This is the second page of stuff")
slider = Slider(option2)
all_screens.append(option2)

# Add the screens to the menu box
option1_button = PushButton(menu, text="Option 1", command=switch_screen, args=[option1], grid=[0,0], align="left")
option2_button = PushButton(menu, text="Option 2", command=switch_screen, args=[option2], grid=[1,0], align="left")

# Hide all screens and then show the first one
hide_all()
all_screens[0].show()

app.display()



from guizero import App, Text, PushButton
from random import randint

a = App("dynamically changing grid", layout="grid")
a.text_size = 40
t = []

def dance():
    for l in t:
        x = l.grid[0] - 1
        if x < 0:
            x = 4

        l.grid = (x, l.grid[1])

for x in range(5):
    for y in range(5):
        t.append(Text(a,
            text="{}.{}".format(x,y),
            grid=[x,y],
            color=(randint(100,255), randint(100,255), randint(100,255))))

a.repeat(250, dance)

a.display()



from guizero import App, Box, Text, TextBox
app = App()

title_box = Box(app, width="fill", align="top", border=True)
Text(title_box, text="title")

buttons_box = Box(app, width="fill", align="bottom", border=True)
Text(buttons_box, text="buttons")

options_box = Box(app, height="fill", align="right", border=True)
Text(options_box, text="options")

content_box = Box(app, align="top", width="fill", border=True)
Text(content_box, text="content")

form_box = Box(content_box, layout="grid", width="fill", align="left", border=True)
Text(form_box, grid=[0,0], text="form", align="right")
Text(form_box, grid=[0,1], text="label", align="left")
TextBox(form_box, grid=[1,1], text="data", width="fill")

app.display()


from guizero import App, Box, Text
from common_test import (
    schedule_after_test,
    schedule_repeat_test,
    destroy_test,
    display_test,
    color_test,
    size_pixel_test,
    size_fill_test,
    events_test,
    cascading_enable_test,
    cascading_properties_test,
    inheriting_properties_test,
    add_tk_widget_test,
    grid_layout_test,
    auto_layout_test
    )

def test_default_values():
    a = App()
    b = Box(a)
    assert b.master == a
    assert b.layout == "auto"
    assert b.grid == None
    assert b.align == None
    assert a.description > ""
    a.destroy()

def test_alt_values():
    a = App(layout = "grid")
    b = Box(a, layout="grid", grid=[0,1], align="top", width=10, height=11)
    assert b.layout == "grid"
    assert b.grid[0] == 0
    assert b.grid[1] == 1
    assert b.align == "top"
    assert b.width == 10
    assert b.height == 11
    a.destroy()

def test_border():
    a = App()
    b = Box(a)
    assert not b.border
    assert b.border == 0

    b.border = True
    assert b.border
    assert b.border == 1
    assert b._get_tk_config("highlightbackground") == "black"

    b.border = False
    assert not b.border

    b.border = 10
    assert b.border
    assert b.border == 10

    b.set_border(11, "red")
    assert b.border
    assert b.border == 11
    assert b._get_tk_config("highlightbackground") == "red"
    a.destroy()


def test_after_schedule():
    a = App()
    b = Box(a)
    schedule_after_test(a, b)
    a.destroy()

def test_repeat_schedule():
    a = App()
    b = Box(a)
    schedule_repeat_test(a, b)
    a.destroy()

def test_destroy():
    a = App()
    b = Box(a)
    destroy_test(b)
    a.destroy()

def test_display():
    a = App()
    b = Box(a)
    display_test(b)
    a.destroy()

def test_color():
    a = App()
    b = Box(a)
    color_test(b)
    a.destroy()

def test_size():
    a = App()
    b = Box(a)
    size_pixel_test(b)
    size_fill_test(b)
    a.destroy()

def test_enable():
    a = App()
    b = Box(a)
    t = Text(b)
    cascading_enable_test(a)
    cascading_enable_test(b)
    a.destroy()

def test_events():
    a = App()
    b = Box(a)
    events_test(b)
    a.destroy()

def test_cascading_properties():
    a = App()
    b = Box(a)
    cascading_properties_test(b)
    a.destroy()

def test_inheriting_properties():
    a = App()
    b = Box(a)
    inheriting_properties_test(b)
    a.destroy()

def test_add_tk_widget():
    a = App()
    add_tk_widget_test(a)
    a.destroy()

def test_auto_layout():
    a = App()
    w = Box(a)
    auto_layout_test(w, None)
    a.destroy()

def test_grid_layout():
    a = App(layout="grid")
    
    w = Box(a, grid=[1,2])
    grid_layout_test(w, 1, 2, 1, 1, None)
    
    ws = Box(a, grid=[1,2,3,4])
    grid_layout_test(ws, 1, 2, 3, 4, None)

    wa = Box(a, grid=[1,2], align="top")
    grid_layout_test(wa, 1, 2, 1, 1, "top")
    
    a.destroy()



from guizero import App, TextBox, PushButton, Text, info
from tkinter.font import Font

def go():
    info("hi", "hi " + textbox.value)

def key_pressed(key):
    print("key pressed {}".format(key))
    print("value = {}".format(textbox.value))

app = App()
text = Text(app, text="Enter your name")
textbox = TextBox(app, width=40)
textbox.update_command(key_pressed)
button = PushButton(app, text="Hi", command=go)
app.display()



from guizero import App, MenuBar

def file_open():
    print("Open")

def file_exit():
    print("Exit")
    app.destroy()

def edit_copy():
    print("Copy")

def edit_cut():
    print("Cut")

def edit_paste():
    print("Paste")

app = App()
menubar = MenuBar(app,
                  toplevel=["File", "Edit"],
                  options=[
                      [ ["Open", file_open], ["Exit", file_exit] ],
                      [ ["Copy", edit_copy], ["Cut", edit_cut], ["Paste", edit_paste] ]
                  ])
app.display()



# form converted from a tkinter example using 'Box'es for layout
# http://www.dreamincode.net/forums/topic/371440-tkinter-overview-with-a-fixed-width-grid/

from guizero import App, MenuBar, Text, TextBox, Combo, Slider, CheckBox, ButtonGroup, Box, PushButton, info, yesno

# Draw the form using 'guizero' wrapper for tkinter
app = App("Simple Window", 400, 300, layout="grid")

# Setup the functions needed by the GUI controls
def exit_function():
    print("File option")
    app.destroy()

def edit_function():
    print("Edit option")

def do_salaryScale(slider_value):
    salaryLabel.value = "Salary : "+str(slider_value)

def do_fullTimeCheckBox():
    if fullTimeCheckBox.value == 1:
        app.title=("Simple Window (full-time)")
    else:
        app.title=("Simple Window")

def do_jobButtonGroup():
    jobTitleLabel.value = jobButtonGroup.value_text

def do_okPushButton():
    info("Information", "Thank you!")

def do_closePushButton():
    if yesno("Exit?", "Exit program?") == True:
        app.destroy()

menubar = MenuBar(app,
                  toplevel=["File"],
                  options=[
                      [ ["Exit", exit_function] ]
                  ])

# Layout the app controls
firstNameLabel = Text(app, text="First Name", grid=[0,0], align="left", size=10)
lastNameLabel = Text(app, text="Last Name", grid=[0,1], align="left", size=10)
countryLabel = Text(app, text="Country", grid=[0,2], align="left", size=10)
addressLabel = Text(app, text="Address", grid=[0,3], align="left", size=10)

firstNameText = TextBox(app, width=25, grid=[1,0])
lastNameText = TextBox(app, width=25, grid=[1,1])
countryCombo = Combo(app, options=['United Kingdom', 'United States', 'France'], selected='United Kingdom', grid=[1,2])
box = Box(app, grid=[1,3], layout="grid")

addressText1 = TextBox(box, width=25, grid=[0,0])
addressText2 = TextBox(box, width=25, grid=[0,1])
addressText3 = TextBox(box, width=25, grid=[0,2])
addressText4 = TextBox(box, width=25, grid=[0,3])
addressText5 = TextBox(box, width=25, grid=[0,4])
addressText6 = TextBox(box, width=25, grid=[0,5])

jobTitleLabel = Text(app, text="TBA", grid=[1,4], size=10)

salaryLabel = Text(app, text="Salary : ", grid=[2,0], align="left", size=10)
salaryScale = Slider(app, start=10000, end=100000, grid=[2,1], command=do_salaryScale, align="left")

fullTimeCheckBox = CheckBox(app, text="Full-time?", grid=[2,2], command=do_fullTimeCheckBox, align="left")
jobButtonGroup = ButtonGroup(app, options=["Programmer", "Developer", "Web Developer", "Designer"],
                             selected=0, grid=[2,3], command=do_jobButtonGroup, align="left")
box1 = Box(app, grid=[2,4], layout="grid", align="left")
okPushButton = PushButton(box1, text="OK", command=do_okPushButton, grid=[0,0], padx=5, pady=5)
closePushButton = PushButton(box1, text="Close", command=do_closePushButton, grid=[2,0], padx=5, pady=5)


app.display()



import pytest
from time import sleep

@pytest.yield_fixture(autouse=True)
def delay():
    # put a small delay in before every test,
    #  otherwise tkinter can get upset by the constant
    #  creating and destroying of apps
    sleep(0.1)


from tkinter import Label, StringVar
from . import utilities as utils
from .base import TextWidget

class Text(TextWidget):

    def __init__(
        self,
        master,
        text="",
        size=None,
        color=None,
        bg=None,
        font=None,
        grid=None,
        align=None,
        visible=True,
        enabled=None,
        width=None,
        height=None):

        self._text = str(text)
        tk = Label(master.tk, text=text)
        super(Text, self).__init__(master, tk, grid, align, visible, enabled, width, height)

        # setup defaults
        if bg:
            self.bg = bg
        
        if size is not None:
            self.text_size = size

        if font is not None:
            self.font = font

        if color is not None:
            self.text_color = color

    # PROPERTIES
    # ----------------------------------
    # The text value
    @property
    def value(self):
        return (self._text)

    @value.setter
    def value(self, value):
        self.tk.config(text=value)
        self._text = str(value)

    # The font size
    @property
    def size(self):
        return self.text_size

    @size.setter
    def size(self, size):
        self.text_size = size

    @property
    def description(self):
        """
        Returns the description for the widget.
        """
        return "[Text] object with text '{}'".format(self.value)

    # METHODS
    # -------------------------------------------

    # Clear text (set to empty string)
    def clear(self):
        self._text = ""
        self.tk.config(text="")

    # Append some text onto the end of the existing text
    def append(self, text):
        new_text = self._text + str(text)
        self._text = new_text
        self.tk.config(text=new_text)




from guizero import App, Window, PushButton, Text

app = App(title="Main window")
Text(app, "Main window")

window = Window(app, title="2nd window", visible=False)
Text(window, "2nd window")

modal_window = Window(app, title="modal window", visible=False)
Text(modal_window, "modal window")

def open_modal():
    modal_window.show(True)

def close_modal():
    modal_window.hide()

open_window_button = PushButton(app, text="Open window", command=window.show)
open_modal_button = PushButton(app, text="Open modal window", command=open_modal)

close_window_button = PushButton(window, text="Close", command=window.hide)

close_modal_button = PushButton(modal_window, text="Close", command=close_modal)

app.display()




from tkinter import messagebox, filedialog
from tkinter.simpledialog import askstring
from . import utilities as utils
import os.path

def warn(title, text, master=None):
    """
    Display a warning message box.

    :param string title:
        The title to be displayed on the box.
    :param string text:
        The text to be displayed on the box.
    :param App master:
        Optional guizero master which the popup will be placed over. Defaults
        to `None`.
    :return:
        None.
    """
    messagebox.showwarning(title, text, parent=None if master is None else master.tk)

def info(title, text, master=None):
    """
    Display an info message box.

    :param string title:
        The title to be displayed on the box.
    :param string text:
        The text to be displayed on the box.
    :param App master:
        Optional guizero master which the popup will be placed over. Defaults
        to `None`.
    :return:
        None.
    """
    messagebox.showinfo(title, text, parent=None if master is None else master.tk)

def error(title, text, master=None):
    """
    Display an error message box.

    :param string title:
        The title to be displayed on the box.
    :param string text:
        The text to be displayed on the box.
    :param App master:
        Optional guizero master which the popup will be placed over. Defaults
        to `None`.
    :return:
        None.
    """
    messagebox.showerror(title, text, parent=None if master is None else master.tk)

def yesno(title, text, master=None):
    """
    Alert type yes or no

    :param string title:
        The title to be displayed on the box.
    :param string text:
        The text to be displayed on the box.
    :param App master:
        Optional guizero master which the popup will be placed over. Defaults
        to `None`.
    :return:
        True if 'yes', False if 'no'
    """
    return messagebox.askyesno(title, text, parent=None if master is None else master.tk)

def question(title, question, initial_value=None, master=None):
    """
    Display a question box which can accept a text response.

    If Ok is pressed the value entered into the box is returned.

    If Cancel is pressed `None` is returned.

    :param string title:
        The title to be displayed on the box.
    :param string text:
        The text to be displayed on the box.
    :param string initial_value:
        The default value for the response box.
    :param App master:
        Optional guizero master which the popup will be placed over. Defaults
        to `None`.
    :return:
        The value entered or `None`.
    """
    return askstring(title, question, initialvalue=initial_value, parent=None if master is None else master.tk)

def select_file(title="Select file", folder=".", filetypes=[["All files", "*.*"]], save=False, master=None):
    """
    Display a box to select a file to open or save.

    If Open or Save is pressed the filename path is returned.

    If Cancel is pressed `None` is returned.

    :param string title:
        The title to be displayed on the box. Defaults to 'Select file'.
    :param string folder:
        The initial folder to open. Defaults to '.'.
    :param list filetypes:
        A list of file descriptions and wildcards of file types to display. 
        Provide as a list of lists e.g. ::
            filetypes=[["Text documents", "*.txt"], ["CSV files", "*.csv"]]
    :param boolean save:
        The default value for the response box.
    :param App master:
        Optional guizero master which the popup will be placed over. Defaults
        to `None`.
    :return:
        The path of filename selected or `None`.
    """
    if not os.path.isdir(folder):
        utils.error_format("The folder '{}' specified for select_file does not exist.".format(folder))
        folder = "."
    
    if save:
        return filedialog.asksaveasfilename(
            title=title, 
            filetypes=filetypes, 
            initialdir=folder, 
            parent=None if master is None else master.tk)
    else:
        return filedialog.askopenfilename(
            title=title, 
            filetypes=filetypes, 
            initialdir=folder, 
            parent=None if master is None else master.tk)

def select_folder(title="Select folder", folder=".", master=None):
    """
    Display a box to select a folder.

    If a folder is selected the folder path is returned, otherwise `None` is returned.

    :param string title:
        The title to be displayed on the box. Defaults to 'Select file'.
    :param string folder:
        The initial folder to open. Defaults to '.'.
    :param App master:
        Optional guizero master which the popup will be placed over. Defaults
        to `None`.
    :return:
        The path of folder selected or `None`.
    """
    if not os.path.isdir(folder):
        utils.error_format("The folder '{}' specified for select_folder does not exist.".format(folder))
        folder = "."
    
    return filedialog.askdirectory(title=title, initialdir=folder, parent=None if master is None else master.tk)


from guizero import App, TextBox
from common_test import (
    schedule_after_test,
    schedule_repeat_test,
    destroy_test,
    enable_test,
    display_test,
    text_test,
    color_test,
    size_text_test,
    size_fill_test,
    events_test,
    cascaded_properties_test,
    inherited_properties_test,
    grid_layout_test,
    auto_layout_test
    )

def test_default_values():
    a = App()
    t = TextBox(a)
    assert t.master == a
    assert t.grid == None
    assert t.align == None
    assert t.width == 10
    assert t.value == ""
    assert a.description > ""
    a.destroy()

def test_alt_values():
    a = App(layout = "grid")
    t = TextBox(
        a,
        text="foo",
        width = 20,
        grid = [0,1],
        align="top",
        multiline=True,
        scrollbar=True)

    assert t.master == a
    assert t.grid[0] == 0
    assert t.grid[1] == 1
    assert t.align == "top"
    assert t.value == "foo\n"
    assert t.width == 20
    a.destroy()

def test_getters_setters():
    a = App()
    t = TextBox(a)
    t.value = "foo"
    assert t.value == "foo"
    a.destroy()

def test_clear():
    a = App()
    t = TextBox(a, text = "foo")
    t.clear()
    assert t.value == ""
    a.destroy()

def test_append():
    a = App()
    t = TextBox(a, text = "foo")
    t.append("bar")
    assert t.value == "foobar"
    a.destroy()

def test_after_schedule():
    a = App()
    t = TextBox(a)
    schedule_after_test(a, t)
    a.destroy()

def test_repeat_schedule():
    a = App()
    t = TextBox(a)
    schedule_repeat_test(a, t)
    a.destroy()

def test_destroy():
    a = App()
    t = TextBox(a)
    destroy_test(t)
    a.destroy()

def test_enable():
    a = App()
    t = TextBox(a)
    enable_test(t)
    a.destroy()

def test_display():
    a = App()
    t = TextBox(a)
    display_test(t)
    a.destroy()

def test_text():
    a = App()
    t = TextBox(a)
    text_test(t)
    a.destroy()

def test_color():
    a = App()
    t = TextBox(a)
    color_test(t)
    a.destroy()

def test_width():
    a = App()
    t = TextBox(a)

    default_width = t.width
    t.width = 30
    assert t.width == 30
    t.width = "fill"
    assert t.width == "fill" 

    a.destroy()

def test_height():
    a = App()

    t1 = TextBox(a)
    t1.height = 10
    #cant change height of single line text box
    assert t1.height == 1

    t2 = TextBox(a, multiline=True)
    default_height = t2.height
    t2.height = 10
    assert t2.height == 10
    t2.height = "fill"
    assert t2.height == "fill"

    a.destroy()

def test_events():
    a = App()
    t = TextBox(a)
    events_test(t)
    a.destroy()

def test_cascaded_properties():
    a = App()
    t = TextBox(a)
    cascaded_properties_test(a, t, True)
    a.destroy()

def test_inherited_properties():
    a = App()
    inherited_properties_test(a, lambda: TextBox(a), True)
    a.destroy()

def test_auto_layout():
    a = App()
    w = TextBox(a)
    auto_layout_test(w, None)
    a.destroy()

def test_grid_layout():
    a = App(layout="grid")
    
    w = TextBox(a, grid=[1,2])
    grid_layout_test(w, 1, 2, 1, 1, None)
    
    ws = TextBox(a, grid=[1,2,3,4])
    grid_layout_test(ws, 1, 2, 3, 4, None)

    wa = TextBox(a, grid=[1,2], align="top")
    grid_layout_test(wa, 1, 2, 1, 1, "top")
    
    a.destroy()

def test_disabled_text():
    a = App()
    t = TextBox(a, enabled=False)
    t.text = "text"
    assert t.text == "text"
    a.destroy()



from tkinter import Label
from . import utilities as utils
from .base import Widget

class Picture(Widget):

    def __init__(
        self, 
        master, 
        image=None, 
        grid=None, 
        align=None, 
        visible=True, 
        enabled=None, 
        width=None, 
        height=None):

        """
        Creates a Picture

        :param Container master:
            The Container (App, Box, etc) the Picture will belong to.

        :param string image:
            A string containing the image to display, defaults to `None`.

        :param List grid:
            Grid co-ordinates for the widget, required if the master layout
            is 'grid', defaults to `None`.

        :param string align:
            How to align the widget within the grid, defaults to None.

        :param bool visible:
            If the widget should be visible, defaults to `True`.

        :param bool enabled:
            If the widget should be enabled, defaults to `None`. If `None`
            the value is inherited from the master.

        :param int width:
            The starting width of the widget. Defaults to `None` and will auto
            size.

        :param int height:
            The starting height of the widget. Defaults to `None` and will auto
            size.
        """

        self._image_source = image
        self._image = None
        self._image_player = None

        # Instantiate label object which will contain image
        tk = Label(master.tk)

        super(Picture, self).__init__(master, tk, grid, align, visible, enabled, width, height)

        # create the image
        self._load_image()

    def _load_image(self):
        if self._height == "fill" or self._width == "fill":
            utils.raise_error("{}\nCannot use 'fill' for width and height.".format(self.description))

        if self._image_source is not None:

            # stop any animation which might still be playing
            if self._image_player:
                self._image_player.stop()

            # load the image and set its properties
            self._image = utils.GUIZeroImage(self._image_source, self._width, self._height)

            self._width = self._image.width
            self._height = self._image.height

            # if its an animation, start it up
            if self._image.animation:
                self._image_player = utils.AnimationPlayer(self, self._image, self._update_tk_image)
            else:
                self._update_tk_image(self._image.tk_image)

            self.tk.config(width=self._width)
            self.tk.config(height=self._height)

    def _update_tk_image(self, tk_image):
        self.tk.config(image=tk_image)

    # PROPERTIES
    # ----------------------------------
    # Get the filename of the image
    @property
    def value(self):
        """
        Gets or sets the file path, tkinter.PhotoImage or PIL.Image you wish to display
        """
        if self._image:
            return self._image.image_source
        else:
            return None

    # Set the image to a given file
    @value.setter
    def value(self, image_source):
        self._image_source = image_source
        self._load_image()

    @property
    def image(self):
        """
        Gets or sets the file path, tkinter.PhotoImage or PIL.Image you wish to display
        """
        return self.value

    # Set the image to a given file
    @image.setter
    def image(self, image_source):
        self.value = image_source

    def resize(self, width, height):
        self._width = width
        self._height = height
        self._load_image()



from guizero import App, Text, TextBox, Slider, PushButton, Combo, Box
from random import randint, choice

colors_text = ["red", "purple", "black", "blue", "green"]
colors_bg = ["pink", "yellow", "orange", "white", "cyan"]
fonts = ["garamond", "arial", "helvetica", "courier new", "times new roman", "comic sans"]

def randomise():
    app.bg = choice(colors_bg)
    for widget in app.children:
        widget.bg = choice(colors_bg)
        widget.text_color = choice(colors_text)
        widget.text_size = randint(8,20)
        widget.font = choice(fonts)

def reset():
    app.bg = None
    app.text_color = None
    app.text_size = None
    app.font = None

app = App(width=300, height=200)

title = Text(app, "Customise your app")
text = TextBox(app, text="then reset to default", width=20)
combo = Combo(app, ["by setting to None"])
code = Text(app, text="widget.bg = None", font="courier new")
box = Box(app, layout="grid")
but_random = PushButton(box, text="random", command=randomise, grid=[0,0])
but_reset = PushButton(box, text="reset", command=reset, grid=[1,0])

app.display()


from guizero import App, Text, TextBox, Combo, PushButton, Box

app = App()

Text(app, text="My form")

form = Box(app, width="fill", layout="grid")
form.border = True

Text(form, text="Title", grid=[0,0], align="right")
TextBox(form, grid=[1,0])

Text(form, text="Name", grid=[0,1], align="right")
TextBox(form, grid=[1,1])

Text(form, text="Age", grid=[0,2], align="right")
TextBox(form, grid=[1,2])

buttons = Box(app, width="fill", align="bottom")

PushButton(buttons, text="Ok", align="left")
PushButton(buttons, text="Cancel", align="left")

app.display()


from guizero import App, Waffle
from tkinter import Canvas, BOTH
import random

# Changes the colour of a randomly chosen pixel to a random colour value
def change_color():
    x = random.randint(1, big_waffle.width-1)
    y = random.randint(1, big_waffle.height-1)
    try:
        rgb = "#%02x%02x%02x" % ( random.randint(0,255), random.randint(0,255), random.randint(0, 255))
        big_waffle.set_pixel(x, y, rgb)
    except:
        print("Oops, something went wrong")

    # Wait 10ms and call this function again
    app.after(10, change_color)


app = App("Waffle!", height=50*20, width=50*20)

# Create a waffle 80x80 with circular pixels 10x10 and 2px padding between them
big_waffle = Waffle(app, height=80, width=80, dim=10, pad=2, dotty=True)

# Call the function once to start it off...
change_color()

app.display()



from tkinter import Listbox, Frame, Scrollbar, END, BROWSE, EXTENDED
from . import utilities as utils
from .base import TextWidget, ContainerTextWidget, Widget
from .event import EventManager


class ListBox(ContainerTextWidget):

    def __init__(
        self,
        master,
        items=None,
        selected=None,
        command=None,
        grid=None,
        align=None,
        visible=True,
        enabled=None,
        multiselect=False,
        scrollbar=False,
        width=None,
        height=None):
        """
        Creates a ListBox

        :param Container master:
            The Container (App, Box, etc) the ListBox will belong too.

        :param List items:
            A list of strings to populate the ListBox, defaults to `None`.

        :param string selected:
            The item in the ListBox to select, defaults to `None`.

        :param Callable command:
            The callback function to call when the ListBox changes,
            defaults to `None`.

        :param List grid:
            Grid co-ordinates for the widget, required if the master layout
            is 'grid', defaults to `None`.

        :param string align:
            How to align the widget within the grid, defaults to None.

        :param bool visible:
            If the widget should be visible, defaults to `True`.

        :param bool enabled:
            If the widget should be enabled, defaults to `None`. If `None`
            the value is inherited from the master.

        :param bool multiselect:
            If ListBox should allow multiple items to be selected, defaults
            to `False`.

        :param bool scrollbar:
            If ListBox should have a vertical scrollbar, defaults to False.

        :param int width:
            The starting width of the widget. Defaults to `None` and will auto
            size.

        :param int height:
            The starting height of the widget. Defaults to `None` and will auto
            size.
        """

        tk = Frame(master.tk)

        super(ListBox, self).__init__(master, tk, "auto", grid, align, visible, enabled, width, height)

        self._listbox = ListBoxWidget(self, items, selected, command, None, "left", visible, enabled, multiselect, None, None)
        self._listbox.resize("fill", "fill")

        if scrollbar:
            # create the scrollbar and add it to the listbox
            scrollbar_tk_widget = Scrollbar(tk)
            ListBoxScrollbar(self, scrollbar_tk_widget, None, "right", True, True, None, "fill")
            self._listbox.tk.config(yscrollcommand=scrollbar_tk_widget.set)
            scrollbar_tk_widget.config(command=self._listbox.tk.yview)

        # override the event manager to associate it to the list box and not the frame
        self._events = EventManager(self, self._listbox.tk)

        self.resize(width, height)

    def resize(self, width, height):
        """
        Resizes the widget.

        :param int width:
            The width of the widget.

        :param int height:
            The height of the widget.
        """
        # set the internal listbox width to be 0 as otherwise it maintains a default
        # size and you cant make the control smaller than that.
        self._listbox._set_tk_config("width", None if width is None else 0)
        self._set_propagation(width, height)
        super(ListBox, self).resize(width, height)

    @property
    def value(self):
        """
        Sets or returns the items selected in a ListBox

        `None` if 0 items are selected.

        If the ListBox is a not `multiselect`, `value` is the item selected.

        If the ListBox is a `multiselect`, `value` is a list of items
        selected.
        """
        return self._listbox.value

    @value.setter
    def value(self, value):
        self._listbox.value = value

    def insert(self, index, item):
        """
        Insert a new `item` at `index`
        """
        self._listbox.insert(index, item)

    def append(self, item):
        """
        Appends a new `item` to the end of the ListBox.
        """
        self._listbox.append(item)

    def remove(self, item):
        """
        Removes the first `item` from the ListBox.

        Returns `True` if an item was removed.
        """
        return self._listbox.remove(item)

    def clear(self):
        """
        Clears all the items in a ListBox
        """
        self._listbox.clear()

    @property
    def items(self):
        """
        Returns a list of items in the ListBox
        """
        return self._listbox.items

    def update_command(self, command):
        """
        Updates the callback command which is called when the ListBox
        changes.

        Setting to `None` stops the callback.

        :param Callable command:
            The callback function to call, it can accept 0 or 1 parameters.

            If it accepts 1 parameter the `value` of the ListBox will be
            passed.
        """
        self._listbox.update_command(command)

class ListBoxWidget(TextWidget):

    def __init__(self, master, items=None, selected=None, command=None, grid=None, align=None, visible=True, enabled=None, multiselect=False, width=None, height=None):

        self._multiselect = multiselect

        # Create a tk OptionMenu object within this object
        mode = EXTENDED if multiselect else BROWSE
        # exportselection=0 allows you to select from more than 1 Listbox
        tk = Listbox(master.tk, selectmode=mode, exportselection=0)

        # Add the items
        if items is not None:
            for item in items:
                tk.insert(END, item)

        super(ListBoxWidget, self).__init__(master, tk, grid, align, visible, enabled, width, height)

        self.events.set_event("<ListBox.ListboxSelect>", "<<ListboxSelect>>", self._command_callback)

        # Select the selected items
        if selected is not None:
            self.value = selected

        # The command associated with this combo
        self.update_command(command)

    @property
    def value(self):
        if len(self.tk.curselection()) > 0:
            if self._multiselect:
                return [self.tk.get(selected) for selected in self.tk.curselection()]
            else:
                return self.tk.get(self.tk.curselection()[0])
        else:
            return None

    @value.setter
    def value(self, value):
        self.tk.selection_clear(0, self.tk.size() - 1)

        # go through all the items and select those in `value`
        for index in range(self.tk.size()):
            if self._multiselect:
                for item in value:
                    if self.tk.get(index) == item:
                        self.tk.select_set(index)
            else:
                if self.tk.get(index) == value:
                    self.tk.select_set(index)

    def insert(self, index, item):
        self.tk.insert(index, item)

    def append(self, item):
        self.tk.insert(END, item)

    def remove(self, item):
        for index in range(len(self.items)):
            if item == self.items[index]:
                self.tk.delete(index)
                return True

        return False

    def clear(self):
        self.tk.delete(0, END)

    @property
    def items(self):
        return [self.tk.get(index) for index in range(self.tk.size())]

    def _command_callback(self):
        if self._command:
            args_expected = utils.no_args_expected(self._command)
            if args_expected == 0:
                self._command()
            elif args_expected == 1:
                self._command(self.value)
            else:
                utils.error_format("ListBox command function must accept either 0 or 1 arguments.\nThe current command has {} arguments.".format(args_expected))

    def update_command(self, command):
        if command is None:
            self._command = lambda: None
        else:
            self._command = command


class ListBoxScrollbar(Widget):
    pass


from guizero import App, Picture

app = App()
picture = Picture(app, image="guizero.gif")
picture_png = Picture(app, image="guizero.png")
picture_jpg = Picture(app, image="guizero.jpg")

app.display()



from guizero import App, Text, Box, Drawing, Slider, PushButton, Combo

def start_paint(event):
    painting.last_event = event

def stop_paint():
    painting.last_event = None
    painting.last_shape = None

def draw_paint(event):
    if tool.value == "line":
        painting.line(
            painting.last_event.x, painting.last_event.y, 
            event.x, event.y, 
            color=(red.value, green.value, blue.value), 
            width=line_width.value)

        painting.last_event = event
    else:
        if painting.last_shape is not None:
            painting.delete(painting.last_shape)
        
        if tool.value == "rectangle":
            painting.last_shape = painting.rectangle(
                painting.last_event.x, painting.last_event.y, 
                event.x, event.y, 
                color=(red.value, green.value, blue.value)
            )
        elif tool.value == "oval":
            painting.last_shape = painting.oval(
                painting.last_event.x, painting.last_event.y, 
                event.x, event.y, 
                color=(red.value, green.value, blue.value)
            )

def update_color():
    color_label.bg = (red.value, green.value, blue.value)
    color_label.text_color = (255 - red.value, 255 - green.value, 255 - blue.value)

def select_tool():
    line_width_box.visible = tool.value == "line"
    
a = App(title="Paintzero")

tool_box = Box(a, height="fill", border=True, align="left")
paint_box = Box(a, width="fill", height="fill", border=True, align="right")

tool = Combo(tool_box, options=["line", "rectangle", "oval"], selected="line", align="top", width="fill", command=select_tool)

color_label = Text(tool_box, text="color", align="top", width="fill")
color_label.bg = "black"
color_label.text_color = "white"

red = Slider(tool_box, end=255, command=update_color)
red.bg = (255, 0, 0)

green = Slider(tool_box, end=255, command=update_color)
green.bg = (0, 255, 0)

blue = Slider(tool_box, end=255, command=update_color)
blue.bg = (0, 0, 255)

line_width_box = Box(tool_box, align="top")
Text(line_width_box, text="width", align="top")
line_width = Slider(line_width_box, start=1, end=10, align="top")

painting = Drawing(paint_box, width="fill", height="fill")
painting.last_event = None
painting.last_shape = None
painting.when_left_button_pressed = start_paint
painting.when_mouse_dragged = draw_paint
painting.when_left_button_released = stop_paint

PushButton(tool_box, text="Clear", command=painting.clear, align="bottom", width="fill")

a.display()


from tkinter import Frame
from . import utilities as utils
from .base import ContainerWidget

class Box(ContainerWidget):

    def __init__(
        self,
        master,
        layout="auto",
        grid=None,
        align=None,
        visible=True,
        enabled=None,
        width=None,
        height=None,
        border=None):
        """
        Creates a Box

        :param Container master:
            The Container (App, Box, etc) the Box will belong too.

        :param string layout:
            The layout the Box should use "auto" or "grid. Defaults to "auto".

        :param List grid:
            Grid co-ordinates for the widget, required if the master layout
            is 'grid', defaults to `None`.

        :param string align:
            How to align the widget within the grid, defaults to `None`.

        :param callback args:
            A list of arguments to pass to the widgets `command`, defaults to
            `None`.

        :param bool visible:
            If the widget should be visible, defaults to `True`.

        :param bool enabled:
            If the widget should be enabled, defaults to `None`. If `None`
            the value is inherited from the master.

        :param int width:
            The starting width of the widget. Defaults to `None` and will auto
            size. If not `None`, both the width and height need to be set.

        :param int height:
            The starting height of the widget. Defaults to `None` and will auto
            size. If not `None`, both the width and height need to be set.

        :param int border:
            Sets the border thickness. `0` or `False` is no border. `True` or 
            value > 1 sets a border. The default is `None`.
        """
        tk = Frame(master.tk)

        super(Box, self).__init__(master, tk, layout, grid, align, visible, enabled, width, height)

        self.resize(width, height)

        if border is not None:
            self.border = border

    @property
    def border(self):
        """
        Sets or returns the border thickness.

        `0` or `False` is no border.
        `True` or value > 1 sets a border

        """
        return self._get_tk_config("highlightthickness")

    @border.setter
    def border(self, value):
        self.set_border(value, "black")

    def set_border(self, thickness, color="black"):
        """
        Sets the border thickness and color.

        :param int thickness:
            The thickenss of the border.

        :param str color:
            The color of the border.
        """
        self._set_tk_config("highlightthickness", thickness)
        self._set_tk_config("highlightbackground", utils.convert_color(color))

    def resize(self, width, height):
        """
        Resizes the widget.

        :param int width:
            The width of the widget.

        :param int height:
            The height of the widget.
        """
        self._set_propagation(width, height)
        
        super(Box, self).resize(width, height)



from tkinter import Entry, StringVar, Text, END
from tkinter.scrolledtext import ScrolledText
from . import utilities as utils
from .base import TextWidget

class TextBox(TextWidget):

    def __init__(
        self,
        master,
        text="",
        width=10,
        height=1,
        grid=None,
        align=None,
        visible=True,
        enabled=None,
        multiline=False,
        scrollbar=False,
        command=None,
        hide_text=False):

        self._multiline = multiline
        
        # Set up controlling string variable
        self._text = StringVar()
        self._text.set( str(text) )

        # Create a tk object for the text box
        if multiline:
            if scrollbar:
                tk = ScrolledText(master.tk, wrap="word")
            else:
                tk = Text(master.tk)
            tk.insert(END,self._text.get())
        else:
            tk = Entry(master.tk, textvariable=self._text)

        super(TextBox, self).__init__(master, tk, grid, align, visible, enabled, width, height)

        self.hide_text = hide_text
        self.update_command(command)

        # Bind the key pressed event
        self.events.set_event("<TextBox.KeyRelease>", "<KeyRelease>", self._key_released)

    # PROPERTIES
    # ----------------------------------
    # The text value
    @property
    def value(self):
        if self._multiline:
            return self.tk.get(1.0,END)
        else:
            return self._text.get()

    @value.setter
    def value(self, value):
        # if the TextBox is disabled, enable it so they value can be set
        was_disabled = not self.enabled
        if was_disabled:
            self.enabled = True

        self._text.set( str(value) )
        if self._multiline:
            self.tk.delete(1.0,END)
            self.tk.insert(END,self._text.get())

        if was_disabled:
            self.enabled = False

    def resize(self, width, height):
        self._width = width
        if width != "fill":
            self._set_tk_config("width", width)

        if height is not None:
            if self._multiline:
                self._height = height
                if height != "fill":
                    self.tk.config(height=height)
            else:
                if isinstance(height, int):
                    if height > 1:
                        utils.error_format("Cannot change the height of a single line TextBox{}".format(self.description))

    @property
    def hide_text(self):
        return self._hide_text

    @hide_text.setter
    def hide_text(self, value):
        self._hide_text = value

        if value == True:
            show_value = "*"
        elif value == False:
            show_value = ""
        else:
            show_value = value
        
        self._set_tk_config("show", show_value)

    @property
    def description(self):
        """
        Returns the description for the widget.
        """
        return "[TextBox] object with text '{}'".format(self.value)

    # METHODS
    # -------------------------------------------
    def _key_released(self, event):
        if self._command:
            args_expected = utils.no_args_expected(self._command)
            if args_expected == 0:
                self._command()
            elif args_expected == 1:
                self._command(event.key)
            else:
                utils.error_format("TextBox command function must accept either 0 or 1 arguments.\nThe current command has {} arguments.".format(args_expected))

    def update_command(self, command):
        if command is None:
            self._command = lambda: None
        else:
            self._command = command

    # Clear text box
    def clear(self):
        self.value = ""

    # Append text
    def append(self, text):
        self.value = self.value + str(text)




from tkinter import Toplevel
from .base import BaseWindow

from . import utilities as utils

class Window(BaseWindow):

    def __init__(
        self, 
        master, 
        title="guizero", 
        width=500, 
        height=500, 
        layout="auto", 
        bg=None, 
        visible=True):

        self._modal = False
        tk = Toplevel(master.tk)

        super(Window, self).__init__(
            master,
            tk,
            title,
            width,
            height,
            layout,
            bg,
            visible)

    def _close_window(self):
        if self._on_close is None:
            self.hide()
        else:
            self._on_close()

    def hide(self):
        """Hide the window."""
        self.tk.withdraw()
        self._visible = False
        if self._modal:
            self.tk.grab_release()

    def show(self, wait = False):
        """Show the window."""
        self.tk.deiconify()
        self._visible = True
        self._modal = wait
        if self._modal:
            self.tk.grab_set()



from guizero import App, Text, PushButton

def show_button_state():
    if button.value:
        button_state.value = "pressed"
    else:
        button_state.value = "released"

app = App()
button = PushButton(app)
button_state = Text(app)
# continually update the button state
app.repeat(50, show_button_state)
app.display()



from guizero import App, PushButton, TextBox

def welcome():
    print("Welcome")

def hi():
    print("Hi")

def message(my_message):
    print(my_message)

def whatever():
    # say whatever using the message function, passing the text as an argument
    app.after(200, message, args=["Whatever"])

def cancel_hi():
    app.cancel(hi)

app = App()

# create some buttons
hi_button = PushButton(app, cancel_hi, text="Stop hi")
what_button = PushButton(app, whatever, text="Whatever")

# after a very short pause, say welcome
app.after(10, welcome)

# keep repeating hi, until it is cancelled
app.repeat(1000, hi)

app.display()



from guizero import App, PushButton, Text, Box

app = App()

top_box = Box(app, align="top", width="fill")
PushButton(top_box, text="to the left", align="left")
PushButton(top_box, text="to the left", align="left")

Text(app, text="everything you own", align="top")

left_box = Box(app, align="left")
left_box.border = True
Text(left_box, text="in a box to the left")

app.display()


# An example showing how to use sensor data, here with the SenseHAT (or SenseHAT emulator)
# Refer to RaspberryPi forums Teaching and learning resources : https://www.raspberrypi.org/forums/viewtopic.php?f=49&t=202386 for some talk using it.
# compass added for show, but commented as it will not work if just enabled

from guizero import *
from sense_hat import SenseHat
# from sense_emu import SenseHat   #use this instead of sense_hat if you don't have a SenseHAT, but are using Raspbian/x86 Desktop

sense=SenseHat()

SENSOR_ENVIRONMENT_UPDATE_FREQUENCY = 1000  #time in milliseconds (ms)
SENSOR_IMU_UPDATE_FREQUENCY = 100

def read_temp_sensor():
    return round(sense.get_temperature(),1)

def read_humidity_sensor():
    return round(sense.get_humidity(),1)

def read_pressure_sensor():
    return round(sense.get_pressure()*0.1,1)


def update_environment_sensors():
    count_text.value=int(count_text.value)+1    #used to debug, i.e. check it is actually incrementing

    temperature_text.value = read_temp_sensor(),"°C"
    humidity_text.value = read_humidity_sensor(),"%"
    pressure_text.value = read_pressure_sensor(),"kPa"


def update_IMU_sensors():
    orient_yaw,orient_pitch,orient_roll = sense.get_orientation().values()
    mag_x,mag_y,mag_z = sense.get_compass_raw().values()
    acc_x,acc_y,acc_z = sense.get_accelerometer_raw().values()
    gyro_x,gyro_y,gyro_z = sense.get_gyroscope_raw().values()

    IMU_orient_yaw_text.value = round(orient_yaw,1)
    IMU_orient_pitch_text.value = round(orient_pitch,1)
    IMU_orient_roll_text.value = round(orient_roll,1)

    IMU_mag_x_text.value = round(mag_x,1)
    IMU_mag_y_text.value = round(mag_y,1)
    IMU_mag_z_text.value = round(mag_z,1)

    IMU_acc_x_text.value = round(acc_x,1)
    IMU_acc_y_text.value = round(acc_y,1)
    IMU_acc_z_text.value = round(acc_z,1)

    IMU_gyro_x_text.value = round(gyro_x,1)
    IMU_gyro_y_text.value = round(gyro_y,1)
    IMU_gyro_z_text.value = round(gyro_z,1)


if __name__ == '__main__':
    app = App(title="Sensor Display!",
              height=230,
              width=420,
              layout='grid')

    title_count = Text(app, "count 'debug':", grid=[0, 0])
    count_text = Text(app, "1", grid=[1, 0])


    title = Text(app, "Temperature Sensor value:", grid=[0, 1])
    temperature_text = Text(app, "xx", grid=[1, 1])

    title2 = Text(app, "Humidity Sensor value:", grid=[0, 2])
    humidity_text = Text(app, "xx", grid=[1, 2])

    title3 = Text(app, "Pressure Sensor value:", grid=[0, 3])
    pressure_text = Text(app, "xx", grid=[1, 3])

    #IMU box

    IMU_title_orient_yaw = Text(app, "Yaw", grid=[1, 5])
    IMU_title_orient_pitch = Text(app, "Pitch", grid=[2, 5])
    IMU_title_orient_roll = Text(app, "Roll", grid=[3, 5])

    IMU_title_orient = Text(app, "Orientation:", grid=[0, 6])
    IMU_orient_yaw_text = Text(app, "xx", grid=[1, 6])
    IMU_orient_pitch_text = Text(app, "xx", grid=[2, 6])
    IMU_orient_roll_text = Text(app, "xx", grid=[3, 6])

    IMU_title_x = Text(app, "X", grid=[1, 8])
    IMU_title_y = Text(app, "Y", grid=[2, 8])
    IMU_title_z = Text(app, "Z", grid=[3, 8])

    IMU_title_mag = Text(app, "Magnetometer µT", grid=[0, 9])
    IMU_title_acc = Text(app, "Accelerometer Gs", grid=[0, 10])
    IMU_title_gyro = Text(app, "Gyroscope rad/s", grid=[0, 11])

    IMU_mag_x_text = Text(app, "xx", grid=[1, 9])
    IMU_mag_y_text = Text(app, "xx", grid=[2, 9])
    IMU_mag_z_text = Text(app, "xx", grid=[3, 9])

    IMU_acc_x_text = Text(app, "xx", grid=[1, 10])
    IMU_acc_y_text = Text(app, "xx", grid=[2, 10])
    IMU_acc_z_text = Text(app, "xx", grid=[3, 10])

    IMU_gyro_x_text = Text(app, "xx", grid=[1, 11])
    IMU_gyro_y_text = Text(app, "xx", grid=[2, 11])
    IMU_gyro_z_text = Text(app, "xx", grid=[3, 11])

    IMU_title_compass = Text(app, "Compass North Bearing", grid=[0, 13])
    IMU_compass_text = Text(app, "xx", grid=[1, 13])

    app.repeat(SENSOR_ENVIRONMENT_UPDATE_FREQUENCY, update_environment_sensors)
    app.repeat(SENSOR_IMU_UPDATE_FREQUENCY, update_IMU_sensors)

    app.display()




"""
Abstract classes for guizero.
"""
from .tkmixins import (
    ScheduleMixin,
    DestroyMixin,
    EnableMixin,
    FocusMixin,
    DisplayMixin,
    TextMixin,
    ColorMixin,
    SizeMixin,
    LayoutMixin,
    EventsMixin)

from . import utilities as utils
from .event import EventManager
from . import dialog
from tkinter import BOTH, X, Y, YES

class Base():

    def __init__(self, tk):
        """
        Base class for all components in guizero.

        :param tk:
            Top level widget instance of Tkinter which usually is the main window of the application

        :return:
            None.
        """
        self._tk = tk
        self._tk_defaults = {}

        # store the tk widgets default keys
        for key in self.tk.keys():
            self._tk_defaults[key] = self.tk[key]

    @property
    def tk(self):
        """
        Returns the tk widget.

        :return:
            tk widget
        """
        return self._tk

    def _has_tk_config(self, key):
        """
        Checks whether the key is configured or not.

        :param key: Key

        :return:
            Keys in tk.keys
        """
        return key in self.tk.keys()

    def _get_tk_config(self, key, default=False):
        """
        Gets the config from the widget's tk object.

        :param string key:
            The tk config key.

        :param bool default:
            Returns the default value for this key. Defaults to `False`.
        """
        if default:
            return self._tk_defaults[key]
        else:
            return self.tk[key]

    def _set_tk_config(self, keys, value):
        """
        Gets the config from the widget's tk object

        :param string/List keys:
            The tk config key or a list of tk keys.

        :param variable value:
            The value to set. If the value is `None`, the config value will be
            reset to its default.
        """

        # if a single key is passed, convert to list
        if isinstance(keys, str):
            keys = [keys]

        # loop through all the keys
        for key in keys:
            if key in self.tk.keys():
                if value is None:
                    # reset to default
                    self.tk[key] = self._tk_defaults[key]
                else:
                    self.tk[key] = value

    def __repr__(self):
        return "guizero.{} object".format(self.__class__.__name__)

class Component(
    Base,
    ScheduleMixin,
    DestroyMixin,
    FocusMixin,
    ColorMixin,
    EventsMixin):

    def __init__(self, master, tk, displayable):
        """
        An abstract class for a component in guizero.
        """
        super(Component, self).__init__(tk)

        self._master = master
        self._events = EventManager(self, tk)
        self._displayable = displayable

        # check the master
        if self.master is not None:
            if isinstance(master, Container):
                self.master._add_child(self)
            else:
                utils.raise_error("{}\nMaster is not an [App], [Window] or [Box]".format(self.description))

    @property
    def master(self):
        """
        Returns the master (container) of this widget, or `None` if it doesn't have one.

        :return:
            Master container of widget, 'None' if it doesn't exists.
        """
        return self._master

    @property
    def description(self):
        """
        Returns the description for the widget.
        """
        return "[{}] object".format(self.__class__.__name__)

    def __str__(self):
        return self.description

    @property
    def events(self):
        """
        Returns the EventManager which can be used to set custom event handlers.
        """
        return self._events

    @property
    def displayable(self):
        """
        Returns whether the Component can be displayed (packed or gridded)

        Note: this is only used by MenuBar and is a candidate for refactoring
        """
        return self._displayable

    def destroy(self):
        """
        Destroy the tk widget.
        """
        # if this widget has a master remove the it from the master
        if self.master is not None:
            self.master._remove_child(self)

        self.tk.destroy()


class Container(Component):

    def __init__(self, master, tk, layout, displayable):
        """
        An abstract class for a container which can hold other widgets.
        """
        super(Container, self).__init__(master, tk, displayable)
        self._children = []
        self._layout_manager = layout
        self._bg = None
        self._text_color = None
        self._text_size = None
        self._font = None
        self._enabled = True

        # inherit from master
        if self.master is not None:
            self.bg = master.bg
            self.text_color = master.text_color
            self.text_size = master.text_size
            self.font = master.font

    @property
    def layout(self):
        """
        Returns the layout type used by this container.
        """
        return self._layout_manager

    @property
    def bg(self):
        """
        Sets or returns the background color of the container.
        """
        return self._bg

    @bg.setter
    def bg(self, value):
        self._bg = utils.convert_color(value)
        super(Container, self.__class__).bg.fset(self, self._bg)
        # cascade bg to child widgets
        for child in self.children:
            if isinstance(child, (Container, Widget)):
                child.bg = self._bg

    @property
    def text_color(self):
        """
        Sets and returns the text color to be used by the widgets
        in the container.

        If set to None (the default) any widgets added to this container
        will use the default text color
        """
        return self._text_color

    @text_color.setter
    def text_color(self, value):
        self._text_color = utils.convert_color(value)
        # cascade text color to child widgets
        for child in self.children:
            if isinstance(child, (Container, TextWidget)):
                child.text_color = self.text_color

    @property
    def text_size(self):
        """
        Sets and returns the text size to be used by the widgets
        in the container.

        If set to None (the default) any widgets added to this container
        will use the default text size
        """
        return self._text_size

    @text_size.setter
    def text_size(self, value):
        self._text_size = value
        # cascade text color to child widgets
        for child in self.children:
            if isinstance(child, (Container, TextWidget)):
                child.text_size = self.text_size

    @property
    def font(self):
        """
        Sets and returns the font to be used by the widgets
        in the container.

        If set to None (the default) any widgets added to this container
        will use the default font
        """
        return self._font

    @font.setter
    def font(self, value):
        self._font = value
        # cascade text color to child widgets
        for child in self.children:
            if isinstance(child, (Container, TextWidget)):
                child.font = self.font

    @property
    def children(self):
        """
        Returns a list of children widgets
        """
        return self._children

    def add_tk_widget(self, tk_widget, grid=None, align=None, visible=True, enabled=None, width=None, height=None):
        """
        Adds a tk widget into a guizero container.

        :param tkinter.Widget tk_widget:
            The Container (App, Box, etc) the tk widget will belong too.

        :param List grid:
            Grid co-ordinates for the widget, required if the master layout
            is 'grid', defaults to `None`.

        :param string align:
            How to align the widget within the grid, defaults to None.

        :param bool visible:
            If the widget should be visible, defaults to `True`.

        :param bool enabled:
            If the widget should be enabled, defaults to `None`. If `None`
            the value is inherited from the master.

        :param int width:
            The starting width of the widget. Defaults to `None` and will auto
            size.

        :param int height:
            The starting height of the widget. Defaults to `None` and will auto
            size.
        """
        # raise a warning if the tk widgets master is not this container
        if self.tk is not tk_widget.master:
            utils.error_format("The tk widget's master is not '{}'.\nIt may not display correctly.".format(self.description))
        return Widget(self, tk_widget, grid, align, visible, enabled, width, height)

    def _add_child(self, child):
        """
        Associates a child widget with this widget.

        Child widgets are used to cascaded to properties (e.g. bg)
        """
        self.children.append(child)

    def _remove_child(self, child):
        """
        Removes a child widgets association with this widget.
        """
        self.children.remove(child)

    def display_widgets(self):
        """
        Displays all the widgets associated with this Container.

        Should be called when the widgets need to be "re-packed/gridded".
        """
        # All widgets are removed and then recreated to ensure the order they
        # were created is the order they are displayed.

        for child in self.children:

            if child.displayable:

                # forget the widget
                if self.layout != "grid":
                    child.tk.pack_forget()
                else:
                    child.tk.grid_forget()

                # display the widget
                if child.visible:
                    if self.layout != "grid":
                        self._pack_widget(child)
                    else:
                        self._grid_widget(child)

    def _pack_widget(self, widget):
        pack_params={}

        if widget.width == "fill" and widget.height == "fill":
            pack_params["fill"] = BOTH
            pack_params["expand"] = YES
        elif widget.width == "fill":
            pack_params["fill"] = X
        elif widget.height == "fill":
            pack_params["fill"] = Y

        if widget.align is not None:
            pack_params["side"] = widget.align

        # this is to cater for scenario's where the frame will not expand to fill the container
        # if aligned - tk weirdness.
        if pack_params.get("side") is None and pack_params.get("fill") == Y:
            pack_params["expand"] = YES

        if pack_params.get("side") in ["top", "bottom"] and pack_params.get("fill") == Y:
            pack_params["expand"] = YES

        if pack_params.get("side") in ["left", "right"] and pack_params.get("fill") == X:
            pack_params["expand"] = YES

        widget.tk.pack(**pack_params)

    def _grid_widget(self, widget):

        grid_params = {
            "column": widget.grid[0],
            "row": widget.grid[1]
        }

        if len(widget.grid) == 4:
            grid_params["columnspan"] = widget.grid[2]
            grid_params["rowspan"] = widget.grid[3]

        if widget.align is not None:
            directions = {"top": "N", "bottom": "S", "left": "W", "right": "E"}
            grid_params["sticky"] = directions[widget.align]

        widget.tk.grid(**grid_params)

    @property
    def enabled(self):
        """
        Sets or Returns the enabled status of this container.

        Setting the property will change the enabled status of all
        widgets in this container
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value:
            self.enable()
        else:
            self.disable()

    def disable(self):
        """
        Disable all the widgets in this container
        """
        self._enabled = False
        for child in self.children:
            if isinstance(child, (Container, Widget)):
                child.disable()

    def enable(self):
        """
        Enable all the widgets in this container
        """
        self._enabled = True
        for child in self.children:
            if isinstance(child, (Container, Widget)):
                child.enable()


class BaseWindow(Container):

    def __init__(self, master, tk, title, width, height, layout, bg, visible):
        """
        Base class for objects which use windows e.g. `App` and `Window`
        """
        super(BaseWindow, self).__init__(master, tk, layout, False)

        # Initial setup
        self.tk.title( str(title) )
        self.tk.geometry(str(width)+"x"+str(height))
        self._on_close = None
        self._full_screen = False

        self.bg = bg

        self.tk.wm_protocol("WM_DELETE_WINDOW", self._close_window)

        self.visible = visible

        self.tk.update()

    # PROPERTIES
    # -----------------------------------

    # The title text
    @property
    def title(self):
        """
        Sets or returns the title displayed in the title bar.
        """
        return self.tk.title()

    @title.setter
    def title(self, text):
        self.tk.title( str(text) )

    # The height of the window
    @property
    def height(self):
        """
        Sets or returns the height of the window
        """
        return self.tk.winfo_height()

    @height.setter
    def height(self, height):
        self.tk.geometry(str(self.tk.winfo_width())+"x"+str(height))
        self.tk.update()

    # The width of the window
    @property
    def width(self):
        """
        Sets or returns the width of the window
        """
        return self.tk.winfo_width()

    @width.setter
    def width(self, width):
        self.tk.geometry(str(width)+"x"+str(self.tk.winfo_height()))
        self.tk.update()

    @property
    def visible(self):
        """
        Sets or returns the visibility of the window
        """
        return self._visible

    @visible.setter
    def visible(self, value):
        if value:
            self.show()
        else:
            self.hide()

    # Whether the window is full screen or not
    @property
    def full_screen(self):
        """
        Sets or returns the visibility of the window
        """
        return self._full_screen

    @full_screen.setter
    def full_screen(self, make_full_screen):
        if make_full_screen:
            self.set_full_screen()
        else:
            self.exit_full_screen()

    @property
    def when_closed(self):
        return self._on_close

    @when_closed.setter
    def when_closed(self, value):
        self._on_close = value

    # METHODS
    # --------------------------------------

    def hide(self):
        """Hide the window."""
        self.tk.withdraw()
        self._visible = False

    def show(self):
        """Show the window."""
        self.tk.deiconify()
        self._visible = True

    def _close_window(self):
        if self._on_close is None:
            self.destroy()
        else:
            self._on_close()

    def update(self):
        self.tk.update()

    def set_full_screen(self, keybind="<Escape>"):
        """Make this window full screen and bind the Escape key (or given key) to exit full screen mode"""
        self.tk.attributes("-fullscreen", True)
        self._full_screen = True
        self.events.set_event("<FullScreen.Escape>", keybind, self.exit_full_screen)

    def exit_full_screen(self):
        """Change from full screen to windowed mode and remove key binding"""
        self.tk.attributes("-fullscreen", False)
        self._full_screen = False
        self.events.remove_event("<FullScreen.Escape>")
    
    def warn(self, title, text):
        dialog.warn(title, text, master=self)

    def info(self, title, text):
        dialog.info(title, text, master=self)

    def error(self, title, text):
        dialog.error(title, text, master=self)

    def yesno(self, title, text):
        return dialog.yesno(title, text, master=self)

    def question(self, title, question, initial_value=None):
        return dialog.question(title, question, initial_value, master=self)

    def select_file(self, title="Select file", folder=".", filetypes=[["All files", "*.*"]], save=False):
        return dialog.select_file(title=title, folder=folder, filetypes=filetypes, save=save, master=self)

    def select_folder(self, title="Select folder", folder="."):
        return dialog.select_folder(title=title, folder=folder, master=self)

    # DEPRECATED METHODS
    # --------------------------------------------
    def on_close(self, command):
        # deprecated on 2019-06-08
        self.when_closed = command
        utils.deprecated("on_close() is deprecated. Please use the when_closed property instead.")


class Widget(
    Component,
    EnableMixin,
    DisplayMixin,
    SizeMixin,
    LayoutMixin):

    def __init__(self, master, tk, grid, align, visible, enabled, width, height):
        """
        The base class for a widget which is an interactable component e.g. `Picture`
        """
        super(Widget, self).__init__(master,tk, True)
        self._update_grid(grid)
        self._update_align(align)
        self._width = width
        self._height = height
        self.visible = visible
        self.enabled = enabled

        # inherit from master
        self.bg = master.bg
        if enabled is None:
            self.enabled = master.enabled
        else:
            self.enabled = enabled

        self.resize(width, height)


class TextWidget(
    Widget,
    TextMixin):

    def __init__(self, master, tk, grid, align, visible, enabled, width, height):
        """
        The base class for a widget which contains or has text e.g. ``Text`, `PushButton`
        """
        super(TextWidget, self).__init__(master, tk, grid, align, visible, enabled, width, height)

        #inherit from master
        self.text_color = master.text_color
        self.text_size = master.text_size
        self.font = master.font


class ContainerWidget(
    Container,
    EnableMixin,
    DisplayMixin,
    SizeMixin,
    LayoutMixin):

    def __init__(self, master, tk, layout, grid, align, visible, enabled, width, height):
        """
        The base class for a widget which is also a container e.g. `Box`, `ButtonGroup`
        """
        super(ContainerWidget, self).__init__(master, tk, layout, True)
        self._update_grid(grid)
        self._update_align(align)
        self._width = width
        self._height = height
        self.visible = visible

        #inherit from master
        if enabled is None:
            self.enabled = master.enabled
        else:
            self.enabled = enabled

    def _set_propagation(self, width, height):
        """
        Set the propagation value of the tk widget dependent on the width and height

        :param int width:
            The width of the widget.

        :param int height:
            The height of the widget.
        """
        if width is None:
            width = 0

        if height is None:
            height = 0

        # set the propagate value
        propagate_function = self.tk.pack_propagate
        if self.layout == "grid":
            propagate_function = self.tk.grid_propagate

        propagate_value = True

        # if height or width > 0 need to stop propagation
        if isinstance(width, int):
            if width > 0:
                propagate_value = False
        if isinstance(height, int):
            if height > 0:
                propagate_value = False

        # if you specify a height or width you must specify they other
        # (unless its a fill)
        if isinstance(width, int) and isinstance(height, int):
            if (width == 0 and height > 0) or (height == 0 and width > 0):
                 utils.error_format("You must specify a width and a height for {}".format(self.description))

        propagate_function(propagate_value)


class ContainerTextWidget(
    ContainerWidget,
    TextMixin):

    def __init__(self, master, tk, layout, grid, align, visible, enabled, width, height):
        """
        The base class for a widget which is also a container and contains text
        e.g. `ButtonGroup`
        """
        super(ContainerTextWidget, self).__init__(master, tk, layout, grid, align, visible, enabled, width, height)



from guizero import App, Text, PushButton

def go_fs():
    app.full_screen = True

def exit_fs():
    app.full_screen = False

app = App()
go_fullscreen = PushButton(app, text="Go full screen", command=go_fs)
exit_fullscreen = Text(app, "To exit full screen, press the escape key")


app.display()



from guizero import App, PushButton, Slider, info

def resize():
    button.width = width.value
    button.height = height.value

app = App()
button = PushButton(app, image="guizero.gif", command=info, args=["button", "you push the button"])

width = Slider(app, start = 10, end = button.width, command = resize)
width.value = button.width

height = Slider(app, start = 10, end = button.height, command = resize)
height.value = button.height

app.display()



from guizero import App, PushButton, Slider

def print_pos(x, y):
    print("{}, {}".format(x, y))

def print_bottom():
    print("bottom button")

def hide():
    b00.hide()
    b01.hide()
    b10.hide()
    b11.hide()
    b_bottom.hide()

def show():
    b00.show()
    b01.show()
    b10.show()
    b11.show()
    b_bottom.show()

app = App(layout="grid")

b00 = PushButton(app, print_pos, text="0,0", args=[0,0], grid=[0,0])
b01 = PushButton(app, print_pos, text="0,1", args=[0,1], grid=[0,1])
b10 = PushButton(app, print_pos, text="1,0", args=[1,0], grid=[1,0])
b11 = PushButton(app, print_pos, text="1,1", args=[1,1], grid=[1,1])

b_bottom = PushButton(app, print_bottom, text="span the bottom", grid=[0,2,2,1])

b_hide = PushButton(app, hide, text="hide", grid=[0,3])
b_show = PushButton(app, show, text="show", grid=[1,3])

app.display()



from guizero import App, PushButton, Text

def button_pressed():
    name = app.question("Hello", "What's your name?")
    # if Cancel is pressed None is return
    # so check a name was entered
    if name is not None:
        hello.value = "Hello " + name

app = App()
button = PushButton(app, command=button_pressed, text="Hello")
hello = Text(app)
app.display()


from tkinter import Checkbutton, IntVar
from . import utilities as utils
from .base import TextWidget

class CheckBox(TextWidget):

    def __init__(
        self,
        master,
        text="",
        command=None,
        grid=None,
        align=None,
        args=None,
        visible=True,
        enabled=None,
        width=None,
        height=None):
        """
        Creates a CheckBox

        :param Container master:
            The Container (App, Box, etc) the CheckBox will belong too.

        :param string selected:
            The text required for the checkbox. Defaults to "".

        :param Callable command:
            The callback function to call when the CheckBox changes.

        :param List grid:
            Grid co-ordinates for the widget, required if the master layout
            is 'grid', defaults to `None`.

        :param string align:
            How to align the widget within the grid, defaults to None.

        :param callback args:
            A list of arguments to pass to the widgets `command`, defaults to
            `None`.

        :param bool visible:
            If the widget should be visible, defaults to `True`.

        :param bool enabled:
            If the widget should be enabled, defaults to `None`. If `None`
            the value is inherited from the master.

        :param int width:
            The starting width of the widget. Defaults to `None` and will auto
            size.

        :param int height:
            The starting height of the widget. Defaults to `None` and will auto
            size.
        """

        self._text = str(text)
        self._value = IntVar()
        tk = Checkbutton(master.tk, text=text, variable=self._value)

        super(CheckBox, self).__init__(master, tk, grid, align, visible, enabled, width, height)

        # Set the command callback
        self.tk.config(command=self._command_callback)
        self.update_command(command, args)

    # PROPERTIES
    # ----------------------------------
    # Whether the box is checked or not
    @property
    def value(self):
        """
        Sets or returns the value of the CheckBox.
        """
        return (self._value.get())

    # Set whether the box is checked or not
    @value.setter
    def value(self, value):
        try:
            value = int(value)
            if value in [0, 1]:
                self._value.set(value)

        except ValueError:
            utils.error_format("You can only set the value of " + self.description + " to either 0 (not checked) or 1 (checked)")

    # The text associated with this box
    @property
    def text(self):
        """
        Sets or returns the text of the CheckBox.
        """
        return (self._text)

    # Set whether the box is checked or not
    @text.setter
    def text(self, text):
        self._text = str(text)
        self.tk.config(text=self._text)

    @property
    def description(self):
        """
        Returns the description for the widget.
        """
        return "[CheckBox] object with text '{}'".format(self.text)

    # METHODS
    # -------------------------------------------

    def toggle(self):
        """
        Toggles the value of the CheckBox.
        """
        self.tk.toggle()

    def update_command(self, command, args=None):
        """
        Updates the callback command which is called when the Combo
        changes.

        Setting to `None` stops the callback.

        :param Callable command:
            The callback function to call.

        :param args list:
            A list of argument values to pass to the callback. Defaults to
            `None`.
        """
        if command is None:
            self._command = lambda: None
        else:
            if args is None:
                self._command = command
            else:
                self._command = utils.with_args(command, *args)

    def _command_callback(self):
        self._command()



from guizero import App, PushButton, info

app = App("A picture button")

picture_button = PushButton(app, command=info, args=["button", "you push the gif"], image="guizero.gif")

picture_button_png = PushButton(app, command=info, args=["button", "you push the png"], image="guizero.png")

picture_button_jpg = PushButton(app, command=info, args=["button", "you push the jpg"], image="guizero.jpg")

app.display()


from guizero import App, Box, Text, PushButton

def hide1():
    t1.hide()

def show1():
    t1.show()

def hide2():
    t2.hide()

def show2():
    t2.show()

def destroy3():
    t3.destroy()

def hide(widget):
    widget.hide()

def show(widget):
    widget.show()

a = App()
b = Box(a, layout="grid")
b.border = True

t1 = Text(a, text="1")
t2 = Text(a, text="2")
t3 = Text(a, text="3")

ta = Text(b, text="aa", grid=[0,0], align="left")
tb = Text(b, text="bbb", grid=[1,0])
tc = Text(b, text="cccc", grid=[2,0])
tc = Text(b, text="ddddd", grid=[0,1])

pbb = Box(a, layout="grid")

b1h = PushButton(pbb, text="hide 1", grid=[0,1], command=hide, args=[t1])
b1s = PushButton(pbb, text="show 1", grid=[1,1], command=show, args=[t1])
b2h = PushButton(pbb, text="hide 2", grid=[2,1], command=hide, args=[t2])
b2s = PushButton(pbb, text="show 2", grid=[3,1], command=show, args=[t2])
bah = PushButton(pbb, text="hide a", grid=[0,0], command=hide, args=[ta])
bas = PushButton(pbb, text="show a", grid=[1,0], command=show, args=[ta])
bbh = PushButton(pbb, text="hide b", grid=[2,0], command=hide, args=[tb])
bbs = PushButton(pbb, text="show b", grid=[3,0], command=show, args=[tb])

a.display()



from guizero import App, Box, Text, PushButton

a = App()

top_box = Box(a, width=500, height=100, align="top")
top_box.border = True

Text(top_box, text="left", align="left")
Text(top_box, text="right", align="right")
Text(top_box, text="top", align="top")
Text(top_box, text="bottom", align="bottom")

a_box = Box(a, width=500, height=100)
a_box.border = True

PushButton(a_box, text="side", align="left")
PushButton(a_box, text="by", align="left")
PushButton(a_box, text="side", align="left")

a_box_in_a_box = Box(a_box, align="right")
a_box_in_a_box.border = True
Text(a_box, text="a box on the right", align="right")
PushButton(a_box_in_a_box, text="on top", align="top")
PushButton(a_box_in_a_box, text="of each other", align="top")

full_button = PushButton(a, text="full button", width="fill", height="fill")

bottom_box = Box(a, width=500, height=100, align="bottom", layout="grid")
bottom_box.border = True

for x in range(16):
    for y in range(4):
        Text(bottom_box, text="{}.{}".format(x,y), grid=[x,y])

Text(a, text="numbers in a grid at the bottom", align="bottom")

a.display()


from guizero import App, Waffle
from threading import Event
from unittest.mock import MagicMock
from common_test import (
    schedule_after_test,
    schedule_repeat_test,
    destroy_test,
    enable_test,
    display_test,
    color_test,
    events_test,
    cascaded_properties_test,
    inherited_properties_test,
    grid_layout_test,
    auto_layout_test
    )

def test_default_values():
    a = App()
    w = Waffle(a)
    assert w.master == a
    assert w.grid == None
    assert w.align == None
    assert w.width == 3
    assert w.height == 3
    assert w.pixel_size == 20
    assert w.pad == 5
    assert w.color == "white"
    assert w.dotty == False
    assert a.description > ""
    a.destroy()

def test_alt_values():
    a = App(layout = "grid")
    w = Waffle(
        a,
        height = 10,
        width = 11,
        dim = 12,
        pad = 13,
        color = "red",
        grid = [0,1],
        align = "top",
        dotty = True,
        bg = "green"
        )
    assert w.master == a
    assert w.grid[0] == 0
    assert w.grid[1] == 1
    assert w.align == "top"
    assert w.height == 10
    assert w.width == 11
    assert w.pixel_size == 12
    assert w.pad == 13
    assert w.color == "red"
    assert w.dotty == True
    assert w.bg == "green"
    a.destroy()

def test_getters_setters():
    a = App()
    w = Waffle(a)

    w.width = 10
    assert w.width == 10
    w.height = 11
    assert w.height == 11
    w.pixel_size = 12
    assert w.pixel_size == 12
    w.pad = 13
    assert w.pad == 13
    w.color = "red"
    assert w.color == "red"
    w.dotty = True
    assert w.dotty == True

    a.destroy()

def test_set_get_all():
    from itertools import chain

    a = App()
    w = Waffle(a)

    w.set_all("red")
    # turn 2d list into 1d list
    pixels = chain.from_iterable(zip(*w.get_all()))

    count = 0
    for pixel in pixels:
        assert pixel == "red"
        count += 1

    assert count == w.width * w.height

    a.destroy()

def test_set_get_pixel():
    a = App()
    w = Waffle(a)

    w.set_pixel(0,0, "red")
    assert w.get_pixel(0,0) == "red"

    a.destroy()

def test_pixel_getters_setters():
    a = App()
    w = Waffle(a)

    pixel = w[0,1]
    assert pixel.x == 0
    assert pixel.y == 1
    assert pixel.size == w.pixel_size
    assert pixel.color == w.color
    assert pixel.dotty == w.dotty

    pixel.color = "red"
    assert pixel.color == "red"

    pixel.dotty = True
    assert pixel.dotty

    a.destroy()

def test_reset():
    from itertools import chain

    a = App()
    w = Waffle(a)

    w.set_pixel(0, 1, "red")
    w.reset()

    pixels = chain.from_iterable(zip(*w.get_all()))

    count = 0
    for pixel in pixels:
        assert pixel == w.color
        count += 1

    assert count == w.width * w.height

    a.destroy()

def test_command():
    a = App()

    callback_event = Event()
    def callback():
        callback_event.set()

    w = Waffle(a, command = callback)
    assert not callback_event.is_set()

    mock_waffle_clicked(w)

    assert callback_event.is_set()

    a.destroy()

def test_command_with_parameters():
    a = App()

    callback_event = Event()
    def callback(x, y):
        assert x == 0
        assert y == 1
        callback_event.set()

    w = Waffle(a, command = callback)
    assert not callback_event.is_set()

    mock_waffle_clicked(w)

    assert callback_event.is_set()

    a.destroy()

def test_update_command():
    a = App()

    callback_event = Event()
    def callback():
        callback_event.set()

    w = Waffle(a)

    mock_waffle_clicked(w)
    assert not callback_event.is_set()

    w.update_command(callback)
    mock_waffle_clicked(w)
    assert callback_event.is_set()
    callback_event.clear()

    w.update_command(None)
    mock_waffle_clicked(w)
    assert not callback_event.is_set()

    a.destroy()

def test_update_command_with_parameters():
    a = App()

    callback_event = Event()
    def callback(x, y):
        assert x == 0
        assert y == 1
        callback_event.set()

    w = Waffle(a)

    w.update_command(callback)

    mock_waffle_clicked(w)
    assert callback_event.is_set()

    a.destroy()

def mock_waffle_clicked(w):
    # you cant invoke a tk canvas - this is better than no tests!
    # create a mock event
    ev = MagicMock()
    ev.widget = w
    ev.tk_event.widget = w._canvas
    # the x event occurs in the middle of the first pixel 
    ev.tk_event.x = ((w.pixel_size + w.pad) / 2)
    # the y event occurs in the middle of the second pixel
    ev.tk_event.y = 1 + (w.pixel_size + w.pad) + ((w.pixel_size + w.pad) / 2)
    w._clicked_on(ev)

def test_after_schedule():
    a = App()
    w = Waffle(a)
    schedule_after_test(a, w)
    a.destroy()

def test_repeat_schedule():
    a = App()
    w = Waffle(a)
    schedule_repeat_test(a, w)
    a.destroy()

def test_destroy():
    a = App()
    w = Waffle(a)
    destroy_test(w)
    a.destroy()

def test_enable():
    a = App()
    w = Waffle(a)
    enable_test(w)
    a.destroy()

def test_display():
    a = App()
    w = Waffle(a)
    display_test(w)
    a.destroy()

def test_color():
    a = App()
    w = Waffle(a)
    color_test(w)
    a.destroy()

def test_events():
    a = App()
    w = Waffle(a)
    events_test(w)
    a.destroy()

def test_cascaded_properties():
    a = App()
    w = Waffle(a)
    cascaded_properties_test(a, w, False)
    a.destroy()

def test_inherited_properties():
    a = App()
    inherited_properties_test(a, lambda: Waffle(a), False)
    a.destroy()

def test_auto_layout():
    a = App()
    w = Waffle(a)
    auto_layout_test(w, None)
    a.destroy()

def test_grid_layout():
    a = App(layout="grid")
    
    w = Waffle(a, grid=[1,2])
    grid_layout_test(w, 1, 2, 1, 1, None)
    
    ws = Waffle(a, grid=[1,2,3,4])
    grid_layout_test(ws, 1, 2, 3, 4, None)

    wa = Waffle(a, grid=[1,2], align="top")
    grid_layout_test(wa, 1, 2, 1, 1, "top")
    
    a.destroy()


from guizero import App, Box, ListBox, CheckBox, Combo, PushButton

app = App()

list_boxes = Box(app, width="fill")
ListBox(list_boxes, items=["Times New Roman", "Courier", "Verdana"], align="left", width="fill")
ListBox(list_boxes, items=["Regular", "Italic", "Bold", "Bold Italic"], align="left", width="fill")
ListBox(list_boxes, items=["8", "10", "12", "14", "16", "18"], align="left", width="fill")

combos = Box(app, width="fill")
combos.border = True
Combo(combos, options=["red", "green", "blue", "yellow"], align="left")
Combo(combos, options=["underline", "double underline"], align="right")

checks = Box(app, width="fill")
checks.border = True
CheckBox(checks, text="strike-through", align="left")
CheckBox(checks, text="double strike-through", align="left")
CheckBox(checks, text="sub-script", align="left")

buttons = Box(app, width="fill", align="bottom")
PushButton(buttons, text="Default", align="left")
PushButton(buttons, text="Ok", align="right")
PushButton(buttons, text="Cancel", align="right")

app.display()



from . import utilities as utils
from tkinter.font import Font
from tkinter import TclError


class ScheduleMixin():
    _callback = {}
    def after(self, time, function, args = []):
        """Call `function` after `time` milliseconds."""
        callback_id = self.tk.after(time, self._call_wrapper, time, function, *args)
        self._callback[function] = [callback_id, False]

    def repeat(self, time, function, args = []):
        """Repeat `function` every `time` milliseconds."""
        callback_id = self.tk.after(time, self._call_wrapper, time, function, *args)
        self._callback[function] = [callback_id, True]

    def cancel(self, function):
        """Cancel the scheduled `function` calls."""
        if function in self._callback.keys():
            callback_id = self._callback[function][0]
            self.tk.after_cancel(callback_id)
            self._callback.pop(function)
        else:
            utils.error_format("Could not cancel function - it doesnt exist, it may have already run")

    def _call_wrapper(self, time, function, *args):
        """Fired by tk.after, gets the callback and either executes the function and cancels or repeats"""
        if function in self._callback.keys():
            repeat = self._callback[function][1]
            if repeat:
                # setup the call back again and update the id
                callback_id = self.tk.after(time, self._call_wrapper, time, function, *args)
                self._callback[function][0] = callback_id
            else:
                # remove it from the call back dictionary
                self._callback.pop(function)
        # execute the function
        function(*args)


class DestroyMixin():
    def destroy(self):
        """Destroy the object."""
        self.tk.destroy()


class EnableMixin():
    @property
    def enabled(self):
        state = self._get_tk_config("state")
        return state == "normal" or state == "active"

    @enabled.setter
    def enabled(self, value):
        if value:
            self.enable()
        else:
            self.disable()

    def disable(self):
        """Disable the widget."""
        self._set_tk_config("state", "disabled")

    def enable(self):
        """Enable the widget."""
        self._set_tk_config("state", "normal")


class FocusMixin():
    def focus(self):
        """Give focus to the widget."""
        self.tk.focus_set()


class DisplayMixin():

    @property
    def visible(self):
        """
        Sets or returns whether the widget is visible.
        """
        return self._visible

    @visible.setter
    def visible(self, value):
        if value:
            self.show()
        else:
            self.hide()

    def hide(self):
        """Hide the widget."""
        self._visible = False
        self.master.display_widgets()

    def show(self):
        """Show the widget."""
        self._visible = True
        self.master.display_widgets()


class TextMixin():

    FG_KEYS = [
        "fg",
        "activeforeground",
    ]

    # Get the font object for this widget
    def _get_font(self, default = False):
        # get the font in use for the widget
        f = Font(self.tk, self._get_tk_config("font", default=default))
        # configure() returns a dictionary of font attributes
        return f.configure()

    # Get the text colour as a string
    @property
    def text_color(self):
        """
        Sets or returns the text color of the widget.
        """
        return self._get_tk_config("fg")

    # Set the text colour
    @text_color.setter
    def text_color(self, color):
        self._set_tk_config(self.FG_KEYS, utils.convert_color(color))

    # Get the current font as a string
    @property
    def font(self):
        """
        Set or returns the font the widget is using.
        """
        f = self._get_font()
        return (f["family"])

    # Set the current font
    @font.setter
    def font(self, font):
        if font is None:
            # get the font from the default font
            f = self._get_font(default = True)
            font = f["family"]

        self._set_tk_config("font", (font, self.text_size))

    # Get the current text size as an integer
    @property
    def text_size(self):
        """
        Sets or returns the text size of the widget.
        """
        f = self._get_font()
        return (f["size"])

    # Set the font size
    @text_size.setter
    def text_size(self, size):
        if size is None:
            # get the size from the default font
            f = self._get_font(default = True)
            size = f["size"]

        self._set_tk_config("font", (self.font, size))


class ColorMixin():

    # these are the widget keys which will get set when the background is changed
    BG_KEYS = [
        "bg",
        "activebackground",
        "selectcolor",
        "troughcolor",
    ]

    # Get the background colour as a string
    @property
    def bg(self):
        """
        Sets the background color of the widget.
        """
        return self._get_tk_config("bg")

    # Set the background colour
    @bg.setter
    def bg(self, color):
        self._set_tk_config(self.BG_KEYS, utils.convert_color(color))


class SizeMixin():
    @property
    def width(self):
        """
        Sets or returns the width of the widget.
        """
        return self._width

    @width.setter
    def width(self, value):
        self.resize(value, self._height)

    @property
    def height(self):
        """
        Sets or returns the height of the widget.
        """
        return self._height

    @height.setter
    def height(self, value):
        self.resize(self._width, value)

    def resize(self, width, height):
        """
        Resizes the widget.

        :param int width:
            The width of the widget.

        :param int height:
            The height of the widget.
        """
        self._width = width
        self._height = height
        if width != "fill":
            self._set_tk_config("width", width)
        if height != "fill":
            self._set_tk_config("height", height)
        if width == "fill" or height == "fill":
            self.master.display_widgets()


class LayoutMixin():
    @property
    def grid(self):
        """
        Sets or returns `[x,y]` coordinates of this widget.
        """
        return self._grid

    @grid.setter
    def grid(self, value):
        self._update_grid(value)
        self.master.display_widgets()

    def _update_grid(self, grid):
        """
        Validates a Widgets grid property and stores it as a TriggeredList
        which will call the masters display_widgets method when it is changed
        """
        self._grid = None
        if self.master.layout == "grid":
            # validate the grid
            if grid is None:
                utils.error_format("{} will not be displayed because it has a missing grid reference.".format(self.description))
            elif not isinstance(grid, (list, tuple)):
                utils.error_format("{} will not be displayed because the grid reference is not a list or tuple.".format(self.description))
            # Can have 2 values (just coords) or 4 values (coords and col/rowspan)
            elif (len(grid) != 2 and len(grid) != 4):
                utils.error_format("{} will not be displayed because the grid reference should be either grid=[x, y] or grid=[x, y, columnspan, rowspan].".format(self.description))
            else:
                # convert the grid to a trackable list
                self._grid = utils.TriggeredList(grid, on_change=self.master.display_widgets)
        else:
            if grid is not None:
                utils.error_format("A grid is not required for {} as it is not using a 'grid' layout.".format(self.description))

    @property
    def align(self):
        """
        Returns the alignment of this widget within its grid location.
        """
        return self._align

    @align.setter
    def align(self, value):
        self._update_align(value)
        self.master.display_widgets()

    def _update_align(self, align):
        """
        Validates a widgets align property
        """
        self._align = None
        if align is not None:
            if align in ["top", "bottom", "left", "right"]:
                self._align = align
            else:
                utils.error_format("Invalid align value ('{}') for {}\nShould be: top, bottom, left or right".format(
                    align,
                    self.description
                ))


class EventsMixin():

    @property
    def when_clicked(self):
        """
        Sets or returns the function called when the widget is clicked.
        """
        return self.events.get_event("<when_clicked>")

    @when_clicked.setter
    def when_clicked(self, value):
        self.events.set_event("<when_clicked>", "<ButtonPress-1>", value)

    @property
    def when_left_button_pressed(self):
        """
        Sets or returns the function called when the left mouse button is
        pressed.
        """
        return self.events.get_event("<when_left_button_pressed>")

    @when_left_button_pressed.setter
    def when_left_button_pressed(self, value):
        self.events.set_event("<when_left_button_pressed>", "<ButtonPress-1>", value)

    @property
    def when_left_button_released(self):
        """
        Sets or returns the function called when the left mouse button is
        released.
        """
        return self.events.get_event("<when_left_button_released>")

    @when_left_button_released.setter
    def when_left_button_released(self, value):
        self.events.set_event("<when_left_button_released>", "<ButtonRelease-1>", value)

    @property
    def when_right_button_pressed(self):
        """
        Sets or returns the function called when the right mouse button is
        pressed.
        """
        return self.events.get_event("<when_right_button_pressed>")

    @when_right_button_pressed.setter
    def when_right_button_pressed(self, value):
        self.events.set_event("<when_right_button_pressed>", "<ButtonPress-3>", value)

    @property
    def when_right_button_released(self):
        """
        Sets or returns the function called when the right mouse button is
        released.
        """
        return self.events.get_event("<when_right_button_released>")

    @when_right_button_released.setter
    def when_right_button_released(self, value):
        self.events.set_event("<when_right_button_released>", "<ButtonRelease-3>", value)

    @property
    def when_key_pressed(self):
        """
        Sets or returns the function called when a key is pressed.
        """
        return self.events.get_event("<when_key_pressed>")

    @when_key_pressed.setter
    def when_key_pressed(self, value):
        self.events.set_event("<when_key_pressed>", "<KeyPress>", value)

    @property
    def when_key_released(self):
        """
        Sets or returns the function called when a key is released.
        """
        return self.events.get_event("<when_key_released>")

    @when_key_released.setter
    def when_key_released(self, value):
        self.events.set_event("<when_key_released>", "<KeyRelease>", value)

    @property
    def when_mouse_enters(self):
        """
        Sets or returns the function called when the mouse pointer enters
        the widget.
        """
        return self.events.get_event("<when_mouse_enters>")

    @when_mouse_enters.setter
    def when_mouse_enters(self, value):
        self.events.set_event("<when_mouse_enters>", "<Enter>", value)

    @property
    def when_mouse_leaves(self):
        """
        Sets or returns the function called when the mouse pointer leaves
        the widget.
        """
        return self.events.get_event("<when_mouse_leaves>")

    @when_mouse_leaves.setter
    def when_mouse_leaves(self, value):
        self.events.set_event("<when_mouse_leaves>", "<Leave>", value)

    @property
    def when_mouse_moved(self):
        """
        Sets or returns the function called when the mouse pointers moves.
        """
        return self.events.get_event("<when_mouse_moved>")

    @when_mouse_moved.setter
    def when_mouse_moved(self, value):
        self.events.set_event("<when_mouse_moved>", "<Motion>", value)

    @property
    def when_mouse_dragged(self):
        """
        Sets or returns the function called when the mouse pointer is dragged
        (moved with the left button pressed).
        """
        return self.events.get_event("<when_mouse_dragged>")

    @when_mouse_dragged.setter
    def when_mouse_dragged(self, value):
        self.events.set_event("<when_mouse_dragged>", "<B1-Motion>", value)



from guizero import App, Text
from common_test import (
    schedule_after_test,
    schedule_repeat_test,
    destroy_test,
    enable_test,
    display_test,
    text_test,
    color_test,
    size_text_test,
    size_fill_test,
    events_test,
    cascaded_properties_test,
    inherited_properties_test,
    SET_FONT,
    TEST_FONTS,
    grid_layout_test,
    auto_layout_test
    )

def test_default_values():
    a = App()
    t = Text(a)
    assert t.master == a
    assert t.grid == None
    assert t.align == None
    assert t.size == 12
    assert t.text_color == "black"
    # test for different fonts to support tests on windows, debian and macos
    assert (t.font == "Arial" or t.font == "Nimbus Sans L" or t.font == "Helvetica")
    assert t.value == ""
    assert a.description > ""
    a.destroy()

def test_alt_values():
    a = App(layout = "grid")
    t = Text(
        a,
        text="foo",
        size = 14,
        color="green",
        bg="red",
        font=SET_FONT,
        grid = [0,1],
        align="top",
        width = 10,
        height = 11)

    assert t.master == a
    assert t.grid[0] == 0
    assert t.grid[1] == 1
    assert t.align == "top"
    assert t.size == 14
    assert t.text_color == "green"
    assert t.bg == "red"
    assert t.font in TEST_FONTS
    assert t.value == "foo"
    assert t.width == 10
    assert t.height == 11
    a.destroy()

def test_getters_setters():
    a = App()
    t = Text(a)
    t.value = "foo"
    assert t.value == "foo"
    t.size = 18
    assert t.size == 18
    a.destroy()

def test_clear():
    a = App()
    t = Text(a, text = "foo")
    t.clear()
    assert t.value == ""
    a.destroy()

def test_append():
    a = App()
    t = Text(a, text = "foo")
    t.append("bar")
    assert t.value == "foobar"
    a.destroy()

def test_after_schedule():
    a = App()
    t = Text(a)
    schedule_after_test(a, t)
    a.destroy()

def test_repeat_schedule():
    a = App()
    t = Text(a)
    schedule_repeat_test(a, t)
    a.destroy()

def test_destroy():
    a = App()
    t = Text(a)
    destroy_test(t)
    a.destroy()

def test_enable():
    a = App()
    t = Text(a)
    enable_test(t)
    a.destroy()

def test_display():
    a = App()
    t = Text(a)
    display_test(t)
    a.destroy()

def test_text():
    a = App()
    # default values
    t = Text(a, color=None, size=None, font=None)
    text_test(t)
    a.destroy()

def test_color():
    a = App()
    t = Text(a)
    color_test(t)
    a.destroy()

def test_size():
    a = App()
    t = Text(a)
    size_text_test(t)
    size_fill_test(t)
    a.destroy()

def test_events():
    a = App()
    t = Text(a)
    events_test(t)
    a.destroy()

def test_cascaded_properties():
    a = App()
    t = Text(a)
    cascaded_properties_test(a, t, True)
    a.destroy()

def test_inherited_properties():
    a = App()
    t = Text(a)
    inherited_properties_test(a, lambda: Text(a), True)
    a.destroy()

def test_auto_layout():
    a = App()
    w = Text(a)
    auto_layout_test(w, None)
    a.destroy()

def test_grid_layout():
    a = App(layout="grid")
    
    w = Text(a, grid=[1,2])
    grid_layout_test(w, 1, 2, 1, 1, None)
    
    ws = Text(a, grid=[1,2,3,4])
    grid_layout_test(ws, 1, 2, 3, 4, None)

    wa = Text(a, grid=[1,2], align="top")
    grid_layout_test(wa, 1, 2, 1, 1, "top")
    
    a.destroy()


from guizero import App, Text

a = App()
a.font = "courier new"

t1 = Text(a)
t1.value = "{}, {}, {}".format(t1.font, t1.text_size, t1.text_color)

t2 = Text(a, font="arial")
t2.value = "{}, {}, {}".format(t2.font, t2.text_size, t2.text_color)

t3 = Text(a, color="red", size=8, font="verdana")
t3.value = "{}, {}, {}".format(t3.font, t3.text_size, t3.text_color)
a.display()


from guizero import App, TextBox, Text, Slider, PushButton, Picture, Combo, CheckBox, ButtonGroup, Box

app = App(title="different sizes", width = 700, height=700)

text = Text(app, "lets change some sizes", width=20, height=2)

text_box = TextBox(app, "some text", width=50)

slider = Slider(app, width=300, height=30)

button = PushButton(app, width=20, height=2)

pic = Picture(app, image="guizero.gif", width=400, height=50)

combo = Combo(app, ["martin", "laura", "rik"], width="fill", height="fill")

check = CheckBox(app, "tick me", width=20, height=3)
check.bg = "blue"

button_group = ButtonGroup(app, ["cheese", "onion", "crisps"], 1, width=20, height=9)
button_group.bg = "darkgrey"

box = Box(app, width=100, height=100)
box.border = True

box.bg = "red"

app.display()


from guizero import App, PushButton
def do_start():
    print("Start button was pressed")
    startbutton.toggle()
    stopbutton.toggle()

def do_stop():
    print("Stop button was pressed")
    startbutton.toggle()
    stopbutton.toggle()

app = App()
startbutton = PushButton(app, command=do_start, text='Start')
stopbutton = PushButton(app, command=do_stop, text='Stop')
stopbutton.disable()
app.display()


from guizero import App, Text, PushButton, ButtonGroup, Picture, info, warn

def change_message(message):
    message_text.value = message

# these functions are called in order based on the users actions
def message1():
    change_message("guizero makes creating GUI's easy")
    app.after(2000, message2)

def message2():
    message = "guizero uses events to make things happen\n"
    message += "events are actions, like clicking a push button\n"
    message += "so...  create an event!"
    change_message(message)
    app.after(1000, show_the_button)

def show_the_button():
    button.show()

def clicked_the_button():
    button.hide()
    change_message("well done, you created an event by clicking the button")
    app.after(2000, message3)

def message3():
    message = "events are any action the user takes in your app\n"
    message += "choose carefully"
    change_message(message)
    app.after(1000, show_the_selection)

def show_the_selection():
    selection.show()

def selection_chosen():
    selection.hide()
    if selection.value == "choose me":
        info("well done", "you chose wisely")
        goodbye()
    else:
        warn("arrrgh", "next time, choose wisely!")
        app.after(2000, show_the_selection)

def goodbye():
    change_message("Thanks")
    logo.show()

# create the app and the widgets
app = App()
message_text = Text(app, text = "Welcome to guizero")
button = PushButton(app, clicked_the_button, text = "Click me")
button.hide()
selection = ButtonGroup(app, selected = 0, command = selection_chosen, options = ["choose me", "don't choose me"])
selection.hide()
logo = Picture(app, "guizero.gif")
logo.hide()

# show the first message after 2 seconds
app.after(2000, message1)

app.display()



from guizero import App, Picture, Slider

def resize():
    picture.width = width.value
    picture.height = height.value

app = App()
picture = Picture(app, image="guizero.gif")

width = Slider(app, start = 10, end = picture.width, command = resize)
width.value = picture.width

height = Slider(app, start = 10, end = picture.height, command = resize)
height.value = picture.height

app.display()



from guizero import App, Waffle, Text, PushButton, info
import random

# Set up the game - colours, width and height of board and no of moves allowed
colours = ["red", "blue", "green", "yellow", "fuchsia", "purple"]
b_width = 14
b_height = 14
moves_limit = 25


# Set up the palette
def init_palette():
    [palette.set_pixel(colours.index(c), 0, c) for c in colours]

# Fill the board with coloured regions
def fill_board():
    [board.set_pixel(x, y, random.choice(colours)) for y in range(b_height) for x in range(b_width)]

# Find and flood any squares next to this
def begin_flood(x, y):
    replacement = palette.get_pixel(x,y)
    target = board.get_pixel(0,0)
    flood(0, 0, target, replacement)
    win_check()

# Recursively floods adjacent squares
def flood(x, y, target, replacement):
    # Algorithm from https://en.wikipedia.org/wiki/Flood_fill
    if target == replacement:
        return False
    if board.get_pixel(x, y) != target:
        return False
    board.set_pixel(x, y, replacement)
    if y+1 <= b_height-1:   # South
        flood(x, y+1, target, replacement)
    if y-1 >= 0:            # North
        flood(x, y-1, target, replacement)
    if x+1 <= b_width-1:    # East
        flood(x+1, y, target, replacement)
    if x-1 >= 0:            # West
        flood(x-1, y, target, replacement)

# Check if there are any moves left or if they won
def win_check():
    moves_left = int(moves_text.value)-1
    moves_text.value = moves_left   # Update moves left
    if moves_left > 0:
        squares = board.get_all()
        if all(colour == squares[0] for colour in squares):
            win_text.value = "Winner!"
            reset.visible = True
            palette.disable()
    else:
        win_text.value = "No more moves left!"
        reset.visible = True
        palette.disable()

# Reset the board and remove the win text/reset button
def reset_board():
    reset.visible = False
    win_text.value = ""
    moves_text.value = moves_limit
    init_palette()
    fill_board()
    palette.enable()



# Set up the game board
app = App("Flood it")

board = Waffle(app, width=b_width, height=b_width, pad=0)
palette = Waffle(app, width=len(colours), height=1, command=begin_flood, dotty=True)
moves_left = Text(app, text="Moves left:")
moves_text = Text(app, text=moves_limit)

# Win text and reset button (initially invisible)
win_text = Text(app)
reset = PushButton(app, text="Start again", command=reset_board)
reset.visible = False

# Initialise the palette and the random board pattern
init_palette()
fill_board()

# Instructions
instructions = PushButton(app, command=info, args=["Instructions", "Click a dot to flood the grid with that colour, beginning from the top left square. You have 25 moves to flood all squares on the grid with the same colour."], text="Instructions")

app.display()



import pytest
from guizero import App, Picture, system_config
from common_test import (
    schedule_after_test,
    schedule_repeat_test,
    destroy_test,
    enable_test,
    display_test,
    text_test,
    color_test,
    events_test,
    cascaded_properties_test,
    inherited_properties_test,
    grid_layout_test,
    auto_layout_test
    )

def test_default_values():
    a = App()
    p = Picture(a)
    assert p.master == a
    assert p.image == None
    assert p.grid == None
    assert p.align == None
    assert a.description > ""
    a.destroy()

def test_alt_values():
    a = App(layout = "grid")
    p = Picture(
        a,
        image = "../examples/guizero.gif",
        grid = [0,1],
        align = "top",
        width = 10,
        height = 11)
    assert p.image == "../examples/guizero.gif"
    assert p._image.tk_image is not None
    assert p.grid[0] == 0
    assert p.grid[1] == 1
    assert p.align == "top"
    assert p.width == 10
    assert p.height == 11
    a.destroy()

def test_getters_setters():
    a = App()
    p = Picture(a)

    assert p.image == None
    p.image = "../examples/guizero.gif"
    assert p.image == "../examples/guizero.gif"
    assert p._image.tk_image is not None
    p.width = 100
    assert p.width == 100
    p.height = 100
    assert p.height == 100

    a.destroy()

@pytest.mark.skipif(system_config.PIL_available == False,
                    reason="PIL not available")
def test_jpg():
    a = App()
    p = Picture(a, image = "../examples/guizero.jpg")

    assert p.image == "../examples/guizero.jpg"
    assert p._image.tk_image is not None
    assert p._image.pil_image is not None

    a.destroy()

@pytest.mark.skipif(system_config.PIL_available == False,
                    reason="PIL not available")
def test_animated_picture():
    a = App()
    p = Picture(a, image = "../examples/guizero_flash.gif")

    assert p._image.tk_image is not None
    assert p._image.pil_image is not None
    assert p._image.animation
    assert p._image_player.running

    a.destroy()

def test_picture_tkobject():
    from tkinter import PhotoImage

    a = App()
    photo_image = PhotoImage(file="../examples/guizero.gif")
    p = Picture(a, image=photo_image)
    assert p.image == photo_image
    assert p._image.tk_image is not None
    a.destroy()

@pytest.mark.skipif(system_config.PIL_available == False,
                    reason="PIL not available")
def test_picture_pilobject():
    from PIL import Image

    a = App()
    pil_image = Image.open("../examples/guizero.gif")
    p = Picture(a, image=pil_image)
    assert p.image == pil_image
    assert p._image.tk_image is not None
    assert p._image.pil_image is not None
    a.destroy()

def test_after_schedule():
    a = App()
    p = Picture(a)
    schedule_after_test(a, p)
    a.destroy()

def test_repeat_schedule():
    a = App()
    p = Picture(a)
    schedule_repeat_test(a, p)
    a.destroy()

def test_destroy():
    a = App()
    p = Picture(a)
    destroy_test(p)
    a.destroy()

def test_enable():
    a = App()
    p = Picture(a)
    enable_test(p)
    a.destroy()

def test_display():
    a = App()
    p = Picture(a)
    display_test(p)
    a.destroy()

def test_color():
    a = App()
    p = Picture(a)
    color_test(p)
    a.destroy()

def test_events():
    a = App()
    p = Picture(a)
    events_test(p)
    a.destroy()

def test_cascaded_properties():
    a = App()
    p = Picture(a)
    cascaded_properties_test(a, p, False)
    a.destroy()

def test_inherited_properties():
    a = App()
    inherited_properties_test(a, lambda: Picture(a), False)
    a.destroy()

def test_auto_layout():
    a = App()
    w = Picture(a)
    auto_layout_test(w, None)
    a.destroy()

def test_grid_layout():
    a = App(layout="grid")
    
    w = Picture(a, grid=[1,2])
    grid_layout_test(w, 1, 2, 1, 1, None)
    
    ws = Picture(a, grid=[1,2,3,4])
    grid_layout_test(ws, 1, 2, 3, 4, None)

    wa = Picture(a, grid=[1,2], align="top")
    grid_layout_test(wa, 1, 2, 1, 1, "top")
    
    a.destroy()


from guizero import *
import random


def read_sensor():
    return random.randrange(3200, 5310, 10) / 100


def update_label():
    text.value = read_sensor()
    # recursive call
    text.after(1000, update_label)


if __name__ == '__main__':
    app = App(title='Sensor Display!',
              height=100,
              width=200,
              layout='grid')

    title = Text(app, 'Sensor value:', grid=[0, 0])
    text = Text(app, "xx", grid=[1, 0])

    text.after(1000, update_label)
    app.display()



from guizero import App, MenuBar
from threading import Event
from common_test import (
    schedule_after_test,
    schedule_repeat_test,
    destroy_test
    )

def test_initial_values():
    a = App()

    callback_event = Event()
    def callback():
        callback_event.set()

    m = MenuBar(
        a,
        ["foo", "bar"],
        [
            [ ["foo1", callback], ["foo2", callback] ],
            [ ["bar1", callback], ["bar2", callback] ]
        ])

    assert m.master == a

    assert not callback_event.is_set()
    assert a.description > ""

    # menu invoke doesnt work...
    # m.tk.invoke(0)
    # assert callback_event.is_set()

    a.destroy()



from threading import Event
from guizero import App, Drawing
from common_test import (
    schedule_after_test,
    schedule_repeat_test,
    destroy_test,
    enable_test,
    display_test,
    events_test,
    cascaded_properties_test,
    inherited_properties_test,
    grid_layout_test,
    auto_layout_test
    )

def test_default_values():
    a = App()
    d = Drawing(a)
    assert d.master == a
    assert d.grid == None
    assert d.align == None
    assert d.width == 100
    assert d.height == 100
    assert a.description > ""
    a.destroy()

def test_alt_values():
    a = App(layout = "grid")
    d = Drawing(
        a,
        grid = [0,1],
        align = "top",
        width = 10,
        height = 11)

    assert d.grid[0] == 0
    assert d.grid[1] == 1
    assert d.align == "top"
    assert d.width == 10
    assert d.height == 11
    a.destroy()

def test_after_schedule():
    a = App()
    d = Drawing(a)
    schedule_after_test(a, d)
    a.destroy()

def test_repeat_schedule():
    a = App()
    d = Drawing(a)
    schedule_repeat_test(a, d)
    a.destroy()

def test_destroy():
    a = App()
    d = Drawing(a)
    destroy_test(d)
    a.destroy()

def test_enable():
    a = App()
    d = Drawing(a)
    enable_test(d)
    a.destroy()

def test_display():
    a = App()
    d = Drawing(a)
    display_test(d)
    a.destroy()

def test_cascaded_properties():
    a = App()
    d = Drawing(a)
    cascaded_properties_test(a, d, False)
    a.destroy()

def test_inherited_properties():
    a = App()
    inherited_properties_test(a, lambda: Drawing(a), False)
    a.destroy()

def test_line():
    a = App()
    d = Drawing(a)    
    id = d.line(1,2,3,4)
    assert id > 0
    a.destroy()

def test_oval():
    a = App()
    d = Drawing(a)
    id = d.oval(1,2,3,4)
    assert id > 0
    a.destroy()

def test_rectangle():
    a = App()
    d = Drawing(a)
    id = d.rectangle(1,2,3,4)
    assert id > 0
    a.destroy()

def test_polygon():
    a = App()
    d = Drawing(a)
    id = d.polygon(1,2,3,4,5,6,7,8)
    assert id > 0
    a.destroy()

def test_triangle():
    a = App()
    d = Drawing(a)
    id = d.triangle(1,2,3,4,5,6)
    assert id > 0
    a.destroy()

def test_image():
    a = App()
    d = Drawing(a)
    id = d.image(1,2,"../examples/guizero.gif")
    assert id > 0
    a.destroy()

def test_text():
    a = App()
    d = Drawing(a)
    id = d.text(1,2,"foo")
    assert id > 0
    a.destroy()

def test_delete():
    a = App()
    d = Drawing(a)
    id1 = d.line(1,2,3,4)
    id2 = d.oval(1,2,3,4)
    assert len(d.tk.find_all()) == 2
    d.delete(id1)
    assert len(d.tk.find_all()) == 1
    d.delete(id2)
    assert len(d.tk.find_all()) == 0
    a.destroy()

def test_clear():
    a = App()
    d = Drawing(a)
    d.line(1,2,3,4)
    d.oval(1,2,3,4)
    assert len(d.tk.find_all()) == 2
    d.clear()
    assert len(d.tk.find_all()) == 0
    a.destroy()

def test_auto_layout():
    a = App()
    w = Drawing(a)
    auto_layout_test(w, None)
    a.destroy()

def test_grid_layout():
    a = App(layout="grid")
    
    w = Drawing(a, grid=[1,2])
    grid_layout_test(w, 1, 2, 1, 1, None)
    
    ws = Drawing(a, grid=[1,2,3,4])
    grid_layout_test(ws, 1, 2, 3, 4, None)

    wa = Drawing(a, grid=[1,2], align="top")
    grid_layout_test(wa, 1, 2, 1, 1, "top")
    
    a.destroy()


from guizero import *

# it is possible to create multiple apps
# guizero will present a warning as you should use Window()
# to create multi window applications

def check(window):
    print("app {} clicked".format(window))
    print("1 - txt={} ".format(tb1.value))
    print("2 - txt={} ".format(tb2.value))

app1 = App(title="App 1")
tb1 = TextBox(app1, grid=[0, 1])
pb1 = PushButton(app1, command=check, args = [1])

app2 = App(title="App 2")
tb2 = TextBox(app2, grid=[0, 1])
pb2 = PushButton(app2, command=check, args = [2])

app1.display()



from guizero import App, ButtonGroup

def selected():
    print(choice.value + " " + choice2.value)

app = App()
choice = ButtonGroup(app, options=["cheese", "ham", "salad"], command=selected)
# You can use specific values for the button group by passing them as a 2d list.
# choice = ButtonGroup(app, options=[["cheese", "c"], ["ham", "h"], ["salad", "s"]], selected="h", command=selected)

choice2 = ButtonGroup(app, command=selected)
choice2.append("sandwich")
choice2.append("salad")

app.display()



from threading import Event
from time import sleep
from unittest.mock import MagicMock
from guizero import Text, Picture
from tkinter import Spinbox

SET_FONT = "Times New Roman"
TEST_FONTS = ["Times New Roman", "Liberation Serif"]

def schedule_after_test(app, widget):
    callback_event = Event()
    def callback():
        callback_event.set()
    assert not callback_event.is_set()
    widget.after(0, callback)
    # call tk to update the app
    app.tk.update()
    assert callback_event.is_set()
    #widget.cancel(callback)

def schedule_repeat_test(app, widget):
    callback_event = Event()
    def callback():
        callback_event.set()
        widget.cancel(callback)
    assert not callback_event.is_set()
    widget.repeat(0, callback)
    # call tk to update the app
    app.tk.update()
    assert callback_event.is_set()

def destroy_test(widget):
    assert widget.tk.winfo_exists()
    widget.destroy()
    assert not widget.tk.winfo_exists()

def enable_test(widget):
    assert widget.enabled
    widget.enabled = False
    assert not widget.enabled
    widget.enabled = True
    assert widget.enabled
    widget.disable()
    assert not widget.enabled
    widget.enable()
    assert widget.enabled

# doesn't work under pytest, app always returns None from focus_get() run
# direct from Python it works fine.
def focus_test(app, widget):
    app.focus()
    app.tk.update()
    assert app.tk.focus_get() == app.tk

    widget.focus()
    app.tk.update()
    assert app.tk.focus_get() == widget.tk

def display_test(widget):
    assert widget.visible
    widget.visible = False
    assert not widget.visible
    widget.visible = True
    assert widget.visible
    widget.hide()
    assert not widget.visible
    widget.show()
    assert widget.visible

def color_test(widget):
    default = widget.bg
    widget.bg = "red"
    assert widget.bg == "red"
    widget.bg = "#ff0000"
    assert widget.bg == "#ff0000"
    widget.bg = (255, 0, 0)
    assert widget.bg == "#ff0000"
    widget.bg = None
    assert widget.bg == default

def size_pixel_test(widget):
    default = widget.width
    widget.width = 666
    assert widget.width == 666
    widget.width = None
    assert widget.width == default

    default = widget.height
    widget.height = 666
    assert widget.height == 666
    widget.height = None
    assert widget.height == default

def size_text_test(widget):
    default = widget.width
    widget.width = 30
    assert widget.width == 30
    widget.width = None
    assert widget.width == default

    default = widget.height
    widget.height = 10
    assert widget.height == 10
    widget.height = None
    assert widget.height == default

def size_fill_test(widget):
    default = widget.width
    widget.width = "fill"
    assert widget.width == "fill"
    widget.width = None
    assert widget.width == default

    default = widget.height
    widget.height = "fill"
    assert widget.height == "fill"
    widget.height = None
    assert widget.height == default

def text_test(widget):
    default = widget.font
    widget.font = SET_FONT
    assert widget.font in TEST_FONTS
    widget.font = None
    assert widget.font == default

    default = widget.text_color
    widget.text_color = "red"
    assert widget.text_color == "red"
    widget.text_color = "#ff0000"
    assert widget.text_color == "#ff0000"
    widget.text_color = (255, 0, 0)
    assert widget.text_color == "#ff0000"
    widget.text_color = None
    assert widget.text_color == default

    default = widget.text_size
    widget.text_size = 16
    assert widget.text_size == 16
    widget.text_size = None
    assert widget.text_size == default

def events_test(widget):

    events_to_test = (
        ("when_clicked", "<when_clicked>"),
        ("when_left_button_pressed", "<when_left_button_pressed>"),
        ("when_left_button_released", "<when_left_button_released>"),
        ("when_right_button_pressed", "<when_right_button_pressed>"),
        ("when_right_button_released", "<when_right_button_released>"),
        ("when_key_pressed", "<when_key_pressed>"),
        ("when_key_released", "<when_key_released>"),
        ("when_mouse_enters", "<when_mouse_enters>"),
        ("when_mouse_leaves", "<when_mouse_leaves>"),
        ("when_mouse_dragged", "<when_mouse_dragged>"),
    )

    callback_event = Event()
    def callback():
        callback_event.set()

    callback_with_param_event = Event()
    def callback_with_param(e):
        assert e.widget == widget
        assert e.key == "A"
        assert e.x == 1
        assert e.y == 2
        assert e.display_x == 3
        assert e.display_y == 4
        callback_with_param_event.set()

    for event_to_test in events_to_test:
        # set the when_attribute to the callback
        setattr(widget, event_to_test[0], callback)
        assert not callback_event.is_set()
        # mock the event
        mock_event(widget, event_to_test[1], "A", 1, 2, 3, 4)
        assert callback_event.is_set()

        callback_with_param_event.clear()
        # set the when_attribute to the callback with a parameter
        setattr(widget, event_to_test[0], callback_with_param)
        assert not callback_with_param_event.is_set()
        # mock the event
        mock_event(widget, event_to_test[1], "A", 1, 2, 3, 4)
        assert callback_with_param_event.is_set()

        callback_event.clear()
        callback_with_param_event.clear()
        # set the when_attribute to None
        setattr(widget, event_to_test[0], None)
        mock_event(widget, event_to_test[1], "A", 1, 2, 3, 4)
        # make sure its not called
        assert not callback_event.is_set()
        assert not callback_with_param_event.is_set()

def mock_event(widget, ref, key, x, y, display_x, display_y):
    # you cant invoke a tk event so we will mock it
    # create a mock event

    # get the event callback
    event_callback = widget.events._refs[ref]

    # mock a tk event
    tk_event = MagicMock()
    tk_event.char = key
    tk_event.x = x
    tk_event.y = y
    tk_event.x_root = display_x
    tk_event.y_root = display_y

    # call the event callback
    event_callback._event_callback(tk_event)

def cascaded_properties_test(container, widget, text):
    container.bg = "red"
    container.enabled = False
    assert widget.bg == "red"
    assert not widget.enabled

    if text:
        container.text_color = "purple"
        assert widget.text_color == "purple"
        container.text_size = 16
        assert widget.text_size == 16

def inherited_properties_test(container, widget_create, text):
    container.bg = "red"
    container.enabled = False
    if text:
        container.text_color = "purple"
        container.text_size = 16

    w = widget_create()

    assert w.bg == "red"
    assert not container.enabled
    if text:
        assert w.text_color == "purple"
        assert w.text_size == 16

def cascading_enable_test(container):

    def check_children(container, test_value):
        for child in container.children:
            assert child.enabled == test_value

    assert container.enabled
    check_children(container, True)

    container.enabled = False
    assert not container.enabled
    check_children(container, False)

    container.enabled = True
    assert container.enabled
    check_children(container, True)

    container.disable()
    assert not container.enabled
    check_children(container, False)

    container.enable()
    assert container.enabled
    check_children(container, True)

def cascading_properties_test(container):
    t = Text(container, color=None, size=None, font=None)
    p = Picture(container)

    container.bg = "red"
    container.text_color = "purple"
    container.text_size = 16
    container.font = SET_FONT
    container.enabled = False

    assert t.bg == "red"
    assert t.text_color == "purple"
    assert t.text_size == 16
    assert t.font in TEST_FONTS
    assert t.enabled == False
    assert p.bg == "red"
    assert p.enabled == False

    # test that destroying widgets removes them as children
    p.destroy()
    container.bg = "green"
    assert t.bg == "green"

def inheriting_properties_test(container):
    container.bg = "red"
    container.text_color = "purple"
    container.text_size = 16
    container.font = SET_FONT
    container.enabled = False

    t = Text(container, color=None, size=None, font=None)
    assert t.bg == "red"
    assert t.text_color == "purple"
    assert t.text_size == 16
    assert t.font in TEST_FONTS
    assert not t.enabled

    p = Picture(container)
    assert p.bg == "red"
    assert not p.enabled

def full_screen_test(window):
    assert window.full_screen == False
    window.full_screen = True
    assert window.full_screen == True
    window.full_screen = False
    assert window.full_screen == False
    window.set_full_screen()
    assert window.full_screen == True
    window.exit_full_screen()
    assert window.full_screen == False

def add_tk_widget_test(container):
    s = Spinbox(from_=0, to=10)
    sw = container.add_tk_widget(s)
    assert s is not None
    assert sw is not None
    assert sw.grid is None
    assert sw.align is None
    assert sw.enabled
    assert sw.visible
    assert sw.width is None
    assert sw.height is None
    
    s2 = Spinbox(from_=0, to=10)
    sw = container.add_tk_widget(s2, align="left", visible=False, enabled=False, width="fill", height="fill")
    assert s is not None
    assert sw is not None
    assert sw.align == "left"
    assert not sw.enabled
    assert not sw.visible
    assert sw.width == "fill"
    assert sw.height == "fill"

def auto_layout_test(widget, align):
    
    assert widget.master.layout == "auto"
    side = "top" if align is None else align
    assert widget.tk.pack_info()["side"] == side

    widget.align = "left"
    assert widget.tk.pack_info()["side"] == "left"

    widget.align = "right"
    assert widget.tk.pack_info()["side"] == "right"

    widget.align = "top"
    assert widget.tk.pack_info()["side"] == "top"

    widget.align = "bottom"
    assert widget.tk.pack_info()["side"] == "bottom"

def grid_layout_test(widget, x, y, col_span, row_span, align):

    sticky = {"top": "n", "bottom": "s", "left": "w", "right": "e", None: ""}

    # test initial values
    assert widget.master.layout == "grid"
    assert int(widget.tk.grid_info()["column"]) == x
    assert int(widget.tk.grid_info()["row"]) == y
    assert int(widget.tk.grid_info()["columnspan"]) == col_span
    assert int(widget.tk.grid_info()["rowspan"]) == row_span
    assert widget.tk.grid_info()["sticky"] == sticky[align]

    # change grid values
    widget.grid[0] = 3
    widget.grid[1] = 4
    assert int(widget.tk.grid_info()["column"]) == 3
    assert int(widget.tk.grid_info()["row"]) == 4
    assert int(widget.tk.grid_info()["columnspan"]) == col_span
    assert int(widget.tk.grid_info()["rowspan"]) == row_span
    assert widget.tk.grid_info()["sticky"] == sticky[align]

    widget.grid = [5,6]
    assert int(widget.tk.grid_info()["column"]) == 5
    assert int(widget.tk.grid_info()["row"]) == 6
    assert int(widget.tk.grid_info()["columnspan"]) == 1
    assert int(widget.tk.grid_info()["rowspan"]) == 1
    assert widget.tk.grid_info()["sticky"] == sticky[align]

    widget.grid = (7,8)
    assert int(widget.tk.grid_info()["column"]) == 7
    assert int(widget.tk.grid_info()["row"]) == 8
    assert int(widget.tk.grid_info()["columnspan"]) == 1
    assert int(widget.tk.grid_info()["rowspan"]) == 1
    assert widget.tk.grid_info()["sticky"] == sticky[align]

    # change span
    widget.grid = [9, 10, 11, 12]
    assert int(widget.tk.grid_info()["column"]) == 9
    assert int(widget.tk.grid_info()["row"]) == 10
    assert int(widget.tk.grid_info()["columnspan"]) == 11
    assert int(widget.tk.grid_info()["rowspan"]) == 12
    assert widget.tk.grid_info()["sticky"] == sticky[align]

    # change align
    widget.align = "left"
    assert widget.tk.grid_info()["sticky"] == sticky["left"]




from guizero import App, Slider, TextBox
def slider_changed(slider_value):
    textbox.value = slider_value

app = App()
slider = Slider(app, command=slider_changed)
textbox = TextBox(app)
app.display()